/* ftp.c
   File Transfer Protocol support module */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "geturl.h"
#include "nmalloc.h"
#include "ftp.h"
#include "connect.h"
#include "host.h"
#include "http.h"
#include "url.h"
#include "mtch.h"
#include "loop.h"

extern char *ftp_acc, *ftp_pass;
extern int verbose, quiet, debug, always_rest;
extern FILE *lfile;

extern int errno;

char respline[MAX_RESPLINE];


/* Write the response of FTP server in line, consisting of not more
   than maxlen bytes. <CR> and <LF> characters are stripped from the
   line. The lines that have '-' as their fourth characters are
   skipped, until the line without it is found. The line is
   null-terminated. */
uerr_t
ftp_response(int fd, char *line, int maxlen)
{
   int i, res;

   do
   {
      for (i = 0; i < maxlen; i++)
      {
	 res = iread(fd, line + i, 1);
	 /* res is number of bytes read. */
	 if (res == 1)
	 {
	    if (line[i] == '\n')
	    {
	       line[i] = 0;
	       /* Get rid of \r. */
	       if (i > 0 && line[i - 1] == '\r')
		  line[i - 1] = 0; /* He he */
	       break;
	    }
	 }
	 else
	    return FTPRERR;
      }
#ifdef DEBUG
      if (debug)
	 fprintf(lfile, "%s\n", line);
#endif
      if (line[i])
	 return FTPRERR;
   } while (line[3] == '-');
   return FTPOK;
}

/* Returns the malloc-ed FTP request, ending with <CR><LF>. */
char *
ftp_request(char *command, char *value)
{
   char *res;

   res = nmalloc(strlen(command) + 1 + strlen(value) + 2 + 1);
   sprintf(res, "%s %s\r\n", command, value);
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "\n--> %s %s\n", command, value);
#endif
   return res;
}

/* Sends the USER and PASS commands to the server, to control
   connection socket csock. */
uerr_t
ftp_login(int csock, char *acc, char *pass)
{
   uerr_t err;
   char *request;

   /* Get greeting. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPSRVERR;
   /* Send USER username. */
   request = ftp_request("USER", acc);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   /* An unprobable possibility of logging without a password. */
   if (*respline == '2')
      return FTPOK;
   /* Else, only response 3 is appropriate. */
   if (*respline != '3')
      return FTPLOGINC;
   /* Send PASS password. */
   request = ftp_request("PASS", pass);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPLOGINC;
   /* All OK. */
   return FTPOK;
}

/* Binds a port, and sends the appropriate PORT command to the FTP
   server. Use acceptport after RETR, to get the socket of data
   connection. */
uerr_t
ftp_port(int csock)
{
   uerr_t err;
   char *request, *bytes;
   unsigned char *in_addr;
   int port;

   /* Setting port to 0 lets the system choose a free port. */
   port = 0;
   /* Bind the port. */
   err = bindport(&port);
   if (err != BINDOK)
      return err;
   /* Get my INET address. */
   if (!(in_addr = inet_address()))
      return HOSTERR;
   /* Construct the argument of PORT (of the form a,b,c,d,e,f) */
   bytes = nmalloc(6 * 4 + 1);
   sprintf(bytes, "%d,%d,%d,%d,%d,%d", in_addr[0], in_addr[1],
	   in_addr[2], in_addr[3], (port & 0xff00) >> 8, port & 0xff);
   /* Send PORT request. */
   request = ftp_request("PORT", bytes);
   free(bytes);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPPORTERR;
   return FTPOK;
}

/* Sends the TYPE request to the server. */
uerr_t
ftp_type(int csock, int type)
{
   char *request;
   uerr_t err;
   char stype[2];

   /* Construct argument. */
   stype[0] = type;
   stype[1] = 0;
   /* Send TYPE request. */
   request = ftp_request("TYPE", stype);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '2')
      return FTPUNKNOWNTYPE;
   /* All OK. */
   return FTPOK;
}

/* Changes the working directory by issuing a CWD command to the
   server. */
uerr_t
ftp_cwd(int csock, char *dir)
{
   char *request;
   uerr_t err;

   /* Send CWD request. */
   request = ftp_request("CWD", dir);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate respone. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline == '5')
      return FTPNSFOD;
   if (*respline != '2')
      return FTPRERR;
   /* All OK. */
   return FTPOK;
}

/* Sends REST command to the FTP server. */
uerr_t
ftp_rest(int csock, long offset)
{
   char *request;
   uerr_t err;
   static char buf[20]; /* Buffer for the number */

   sprintf(buf, "%ld", offset);
   request = ftp_request("REST", buf);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline != '3')
      return FTPRESTFAIL;
   /* All OK. */
   return FTPOK;
}

/* Sends RETR command to the FTP server. */
uerr_t
ftp_retr(int csock, char *file)
{
   char *request;
   uerr_t err;

   /* Send RETR request. */
   request = ftp_request("RETR", file);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate response. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline == '5')
      return FTPNSFOD;
   if (*respline != '1')
      return FTPRERR;
   /* All OK. */
   return FTPOK;
}

/* Sends the LIST command to the server. */
uerr_t
ftp_list(int csock, char *file)
{
   char *request;
   uerr_t err;

   /* Send LIST request. */
   request = ftp_request("LIST", file);
   write(csock, request, strlen(request));
   free(request);
   /* Get appropriate respone. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
      return err;
   if (*respline == '5')
      return FTPNSFOD;
   if (*respline != '1')
      return FTPRERR;
   /* All OK. */
   return FTPOK;
}

/* Retrieves a file with denoted parameters through opening an FTP
   connection to the server. Control and data socket are closed
   afterwards. */
uerr_t
getftp(char *remote_host, int remote_port, char *remote_file,
       char *local_file, char *user, char *passwd, long *len,
       long restval, int *res, int *cfd, int *type)
{
   int csock, dtsock, list_only = 0;
   uerr_t err;
   char *dir, *file, *buf;
   FILE *fp;

   *type &= ~RETROKF;
   *len = 0;
   /* First, some checking... */
   buf = nstrdup(remote_file); /* Make a copy of remote_file to be
				  able to modify it. */
   parse_dir(buf, &dir, &file);
   if (!strlen(dir))
      dir = "/";
   if (*type & REMOTELOC)
      file = local_file;

   /* Do we have just to list? */
   list_only = (*type & LIST_ONLY || !strlen(file)) ? 1 : 0;
   
   if (*type & LOGGED_IN)
      csock = *cfd;
   else
   {
      /* Then, let's establish the connection. */
      if (verbose)
	 fprintf(lfile, "Connecting to %s:%d... ", remote_host, remote_port);
      err = make_connection(&csock, remote_host, remote_port);
      if (*type & LEAVE_PENDING)
	 *cfd = csock;
      switch (err)
      {
	 /* Do not close the socket in first several cases,
	    since it wasn't made in the first place. */
	 case HOSTERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "No such host %s.\n", remote_host);
	    }
	    free(buf);
	    return HOSTERR;
	    break;
	 case CONSOCKERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "Socket error (%s).\n", mystrerror(errno));
	    }
	    free(buf);
	    return CONSOCKERR;
	    break;
	 case CONTIMEDOUT:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "Connection to %s timed out.\n", remote_host);
	    }
	    free(buf);
	    close(csock);
	    return CONTIMEDOUT;
	    break;
	 case CONREFUSED:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "Connection to %s refused.\n", remote_host);
	    }
	    free(buf);
	    close(csock);
	    return CONREFUSED;
	 case CONERROR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "connect: %s\n", mystrerror(errno));
	    }
	    free(buf);
	    close(csock);
	    return CONERROR;
	    break;
      }
      /* Log in. */
      if (verbose)
      {
	 fprintf(lfile, "Connected!\n");
	 fprintf(lfile, "Logging in as %s ... ", user);
      }
      err = ftp_login(csock, user, passwd);
      /* All FTPRERR, FTPSRVERR and FTPLOGININC. */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPSRVERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "Error in server greeting.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	 case FTPLOGINC:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "Login incorrect or the server refuses to connect.\n");
	    }
	    free(buf);
	    close(csock);
	    return FTPLOGINC;
	    break;
	 case FTPOK:
	    if (verbose)
	       fprintf(lfile, "Logged in!\n");
	    break;
	 default:
	    fprintf(lfile, "This condition should not have been reached.\n");
	    exit(1);
	    break;
      }
      /* Change type to binary. */
      if (verbose)
	 fprintf(lfile, "==> TYPE I ... ");
      err = ftp_type(csock, 'I');
      /* ftp_response + FTPUNKNOWNTYPE. */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPUNKNOWNTYPE:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "Unknown type.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	 default:
	    /* Everything is OK. */
	    break;
      }
      /* Change working directory. */
      if (verbose)
      {
	 fprintf(lfile, "done!  ");
	 fprintf(lfile, "==> CWD %s ... ", dir);
      }
      err = ftp_cwd(csock, dir);
      /* Possible errors: ftp_response + FTPNSFOD. */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPNSFOD:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "No such directory %s.\n", dir);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
      }
   } /* *type & LOGGED_IN */
   /* Send PORT. */
   if (verbose)
   {
      if (!(*type & LOGGED_IN))
	 fprintf(lfile, "done!\n");
      fprintf(lfile, "==> PORT ... ");
   }
   err = ftp_port(csock);
   /* All possible errors from bindport and ftp_response
      plus FTPPORTERR and HOSTERR. */
   switch(err)
   {
      case HOSTERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "No such host %s.\n", remote_host);
	 }
	 free(buf);
	 close(csock);
	 return HOSTERR;
	 break;
      case CONSOCKERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Socket error (%s).\n", mystrerror(errno));
	 }
	 free(buf);
	 close(csock);
	 return CONSOCKERR;
	 break;
      case BINDERR: case CONPORTERR: case LISTENERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Bind error (%s).\n", mystrerror(errno));
	 }
      case FTPPORTERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Invalid PORT (%s).\n", respline);
	 }
      default:
	 /* Everything is OK. */
	 break;
   }
   /* Restart if needed. */
   if (restval && !list_only)
   {
      if (verbose)
      {
	 fprintf(lfile, "done!    ");
	 fprintf(lfile, "==> REST %ld ... ", restval);
      }
      err = ftp_rest(csock, restval);
      /* Possible errors: ftp_response + FTPRESTFAIL. */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPRESTFAIL:
	    if (verbose)
	       fprintf(lfile, "\nREST failed, starting from scratch.\n");
	    restval = 0;
	    break;
      }
   }
   /* Tell the server what to send. */
   if (!list_only)
   {
      if (verbose)
      {
	 if (err != FTPRESTFAIL)
	    fprintf(lfile, "done!    ");
	 if (restval)
	    fprintf(lfile, "\n");
	 fprintf(lfile, "==> RETR %s ... ", file);
      }
      err = ftp_retr(csock, file);
      /* Possible errors: ftp_response + FTPNSFOD. */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPNSFOD:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "No such file %s.\n", file);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
      }
   }
   else /* list_only */
   {
      if (verbose)
      {
	 if (err != FTPRESTFAIL)
	    fprintf(lfile, "done!    ");
	 if (restval)
	    fprintf(lfile, "\n");
	 fprintf(lfile, "==> LIST ... ");
      }
      err = ftp_list(csock, ".");
      /* Possible errors: ftp_response */
      switch (err)
      {
	 case FTPRERR:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile,
		       "Error in server response, control connection closed.\n");
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
	 case FTPNSFOD:
	    if (!quiet)
	    {
	       if (verbose)
		  fprintf(lfile, "\n");
	       fprintf(lfile, "No such file %s.\n", file);
	    }
	    free(buf);
	    close(csock);
	    return err;
	    break;
      }
   } /* !list_only */
   if (verbose)
   {
      fprintf(lfile, "done!");
      if (!(*type & LOGGED_IN))
	 fprintf(lfile, "\n");
   }
   /* Open the data transmission socket by calling acceptport. */
   err = acceptport(&dtsock);
   /* Possible errors: ACCEPTERR. */
   if (err == ACCEPTERR)
   {
      if (!quiet)
	 fprintf(lfile, "accept: %s\n", mystrerror(errno));
      free(buf);
      close(csock);
      return err;
   }
   /* Open the file... */
   fp = fopen(local_file, restval ? "a" : "w");
   if (!fp)
   {
      if (!quiet)
	 fprintf(lfile, "%s: %s\n", local_file, mystrerror(errno));
      free(buf);
      close(csock);
      return FOPENERR;
   }
   /* As soon as we got away with accept, the document is considered
      RETROKF ok. */
   *type |= RETROKF;
   
   /* Get the contents of the document. */
   *res = get_contents(dtsock, fp, len, restval);
   /* Close data connection socket. */
   closeport(dtsock);
   /* Close the local file. */
   fclose(fp);
   /* If get_contents couldn't write to fp, bail out. */
   if (*res == -2)
   {
      close(csock);
      free(buf);
      return FWRITEERR;
   }
   /* Get server to tell you if everything is retrieved. */
   err = ftp_response(csock, respline, MAX_RESPLINE - 1);
   if (err != FTPOK)
   {
      if (!quiet)
      {
	 /* A different message here, since the connection was
            probably lost, and there is no other mysterious error in
            server response. Newline is not printed because of
            possible Data connection error. */
	 fprintf(lfile, "Control connection closed.");
      }
      free(buf);
      close(csock);
      return err;
   }
   /* If retrieval failed for any reason, return FTPRETRINT. */
   if (*respline != '2')
   {
      free(buf);
      close(csock);
      return FTPRETRINT;
   }
   if (!(*type & LEAVE_PENDING))
   {
      /* I should probably send 'QUIT', but this is faster. */
      close(csock);
   }
   /* Now everything is definitely supposed to be OK. */
   if (list_only && !(*type & NOTFTP2HTML))
   {
      char **files, **dirs;
      int i;
      
      err = ftp_parse_list(local_file, &dirs, &files);
      if (err != FOPENERR)
      {
	 err = ftp_index(local_file, dirs, files, remote_host,
			 remote_file, remote_port);
	 if (err != FOPENERR)
	    if (verbose)
	       fprintf(lfile, "Wrote HTML index to %s.\n", local_file);
         for (i = 0; files[i]; i++)
	    free(files[i]);
         free(files);
         for (i = 0; dirs[i]; i++)
	    free(dirs[i]);
         free(dirs);
      }
   }
   return RETRFINISHED;
}

/* The function parses the listing given in file, and returns hash
   arrays of files and directories. */
uerr_t
ftp_parse_list(char *file, char ***dirs, char ***files)
{
   FILE *fp;
   static char *months[] = {
      "Jan", "Feb", "Mar", "Apr", "May", "Jun",
      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
   };
   int c, next, len, i;
   char *line, *tok, ***cur;

   fp = fopen(file, "r");
   if (!fp)
      return FOPENERR;
   /* Initialize dirs and files. */
   if (dirs)
      *dirs = nmalloc(sizeof(char **));
   *files = nmalloc(sizeof(char **));
   (*files)[0] = NULL;
   if (dirs)
      (*dirs)[0] = NULL;

   /* Line loop to end of file: */
   do
   {
      line = nmalloc(1);
      line[0] = 0;
      len = 0;
      /* Character loop to end of line: */
      do
      {
	 c = fgetc(fp);
	 /* Exit on EOF or newline. */
	 if (c == EOF || c == '\n')
	    break;
	 /* Increase the line size. */
	 line = nrealloc(line, len + 2);
	 line[len] = c;
	 line[len + 1] = 0;
	 ++len;
      } while (1);
      /* Destroy <CR> if there is one. */
      if (line[len] == '\r')
	 line[len] = 0;
      /* Skip if total... */
      if (!strncmp(line, "total", 5))
      {
	 free(line);
	 continue;
      }
      /* Get the first token (permissions). */
      tok = strtok(line, " ");
      if (!tok)
      {
	 free(line);
	 continue;
      }
      /* Decide whether we deal with a file or a directory. */
      if (*tok == 'd')
      {
	 if (dirs)
	    cur = dirs;
	 else
	 {
	    free(line);
	    continue;
	 }
      }
      else if (*tok == '-')
	 cur = files;
      else if (*tok == 'l')
      {
	 /* Chuck out the links. */
	 free(line);
	 continue;
      }
      else /* hmmm */
	 cur = files; /* bletch */
      
      next = -1;
      /* While there are tokens on the line: */
      while ((tok = strtok(NULL, " ")))
      {
	 --next;
	 for (i = 0; i < 12; i++)
	    if (!strcmp(tok, months[i]))
	       break;
	 /* If we got a month, it means the filename is three tokens
            away. */
	 if (i != 12)
	    next = 3;
	 /* If we have the filename, add it to the list of files or
            directories. */
	 if (!next)
	 {
	    /* "." and ".." are an exception! */
	    if (!strcmp(tok, ".") || !strcmp(tok, ".."))
	       break;
	    for (i = 0; (*cur)[i]; i++);
	    *cur = nrealloc(*cur, (i + 2)*sizeof(char *));
	    (*cur)[i] = nstrdup(tok);
	    (*cur)[i + 1] = NULL;
	 }
      }
      /* Free the line. */
      free(line);
   } while (c != EOF);
   /* Close the file. */
   fclose(fp);
   return FTPOK;
}

/* The function creates an HTML index containing references to given
   directories and files on the appropriate host. The references are
   FTP. */
uerr_t
ftp_index(char *file, char **dirs, char **files, char *host, char *dir,
		 int port)
{
   FILE *fp;
   int i;
   char *tmp;

   fp = fopen(file, "w");
   if (!fp)
      return FOPENERR;
   fprintf(fp, "<html>\n<head>\n<title>Index of ftp://%s:%d%s</title>\n</head>\n",
           host, port, dir);
   fprintf(fp, "<body>\n<h1>Index of ftp://%s:%d%s</h1>\n<hr>\n",
           host, port, dir);
   fprintf(fp, "<ul>\n");
   for (i = 0; files[i]; i++)
   {
      tmp = nmalloc(6 + strlen(host) + 1 + numdigit(port) + strlen(dir)
		    + strlen(files[i]) + 1);
      sprintf(tmp, "ftp://%s:%d%s%s", host, port, dir, files[i]);
      fprintf(fp, "   <li>File <a href=\"%s\">%s</a>\n", tmp, files[i]);
      free(tmp);
   }
   for (i = 0; dirs[i]; i++)
   {
      tmp = nmalloc(6 + strlen(host) + 1 + numdigit(port) + strlen(dir)
		    + strlen(dirs[i]) + 1 + 1);
      sprintf(tmp, "ftp://%s:%d%s%s/", host, port, dir, dirs[i]);
      fprintf(fp, "   <li>Directory <a href=\"%s\">%s</a>\n", tmp, dirs[i]);
      free(tmp);
   }
   fprintf(fp, "</ul>\n</body>\n</html>\n");
   fclose(fp);
   return FTPOK;
}

/* Similar to retrieve_url, but specialized for FTP, with globbing. I
   tried to incorporate something like this to retrieve_url (which is
   very ugly as it is), but I gave up when I saw what it looked
   like. */
uerr_t
ftp_glob_loop(char *url, int ntry)
{
   char *remote_dir, *remote_host, *tms;
   char *pat, **files, *dir, *tmp_remote_dir, *local_name;
   uerr_t err, proto;
   int where, ftp_csock, dt, remote_port, count, res, i, match;
   long len, restval;
   struct stat st;
   
   /* Prepare memory for parseurl. */
   remote_dir = nmalloc(strlen(url) + 1);
   remote_host = nmalloc(strlen(url) + 1);
   /* Parse the URL. */
   proto = parseurl(url, remote_host, &remote_port, remote_dir);
   if (proto != URLFTP)
   {
      if (!quiet)
	 fprintf(lfile, "Only FTP protocol is appropriate when globbing.\n");
      free(remote_dir);
      free(remote_host);
      return URLUNKNOWN;
   }

   parse_dir(remote_dir, &dir, &pat);
   /* Starting with listing (where = 0), finish when where reaches
      -1. */
   ftp_csock = 0;
   dt = 0;
   files = NULL;
   for (where = 0; 1; where++)
   {
      if (where == 1) /* meaning the list was retrieved */
      {
	 ftp_parse_list(LIST_FILENAME, NULL, &files);
	 unlink(LIST_FILENAME);
      }
      
      if (where && !files[where - 1])
	 break;

      if (where)
      {
	 /* Match the file against the pattern. */
	 match = fnmatch(pat, files[where - 1], 0);
	 if (match == -1 && *pat)
	 {
	    free(remote_dir);
	    free(remote_host);
	    if (!quiet)
	       fprintf(lfile, "%s: Invalid pattern\n", pat);
	    return URLBADPATTERN;
	 }
	 /* Zero-length pattern means take everything. */
	 else if (match == FNM_NOMATCH && *pat)
	    continue;
      }
      
      /* Reset the counter. */
      count = 0;
      /* Start the loop. */
      do
      {
	 /* Increment the pass counter. */
	 ++count;
	 /* Suppose no restarting gets done. */
	 restval = 0;
	 if (where && (count > 1 || always_rest) && exists(files[where - 1]))
	    if (stat(files[where - 1], &st) == 0)
	       restval = st.st_size;
	 /* Get the current time string. */
	 tms = time_str();
	 /* Set local name */
	 local_name = NULL;
	 if (where)
	 {
	    if (!restval)
	       for (i = 0; !(local_name = unique_name(files[where - 1], i));
		    i++);
	    else
	       local_name = nstrdup(files[where - 1]);
	 }
	 /* Print fetch message, if verbose. */
	 if (verbose)
	 {
	    if (!where)
	       fprintf(lfile,
		       "--%s--[%d]: %s (LIST)\n                 \"%s\"\n",
		       tms, count, url, LIST_FILENAME);
	    else /* where != 0 */
	       fprintf(lfile,
		       "\n--%s--[%d]: %s (GET %s)\n                 \"%s\"\n",
		       tms, count, url, files[where - 1], local_name);
	 }

	 /* Call the FTP fetcher. */
	 if (where) /* Retrieving files... */
	 {
	    tmp_remote_dir = nmalloc(strlen(dir) + 1 + strlen(files[where - 1]) + 1);
	    sprintf(tmp_remote_dir, "%s/%s", dir, files[where - 1]);
	    dt |= LEAVE_PENDING;
	    if (ftp_csock)
	       dt |= LOGGED_IN;
	    dt &= ~LIST_ONLY;
	    err = getftp(remote_host, remote_port, tmp_remote_dir,
			 local_name, ftp_acc, ftp_pass, &len,
			 restval, &res, &ftp_csock, &dt);
	    free(tmp_remote_dir);
	 }
	 else /* Retrieving the listing... */
	 {
	    int ln = strlen(dir);
	    /* Add the trailing '/' to the directory name! */
	    tmp_remote_dir = nmalloc(ln + 2);
	    memcpy(tmp_remote_dir, dir, ln + 1);
	    if (tmp_remote_dir[ln - 1] != '/')
	    {
	       tmp_remote_dir[ln] = '/';
	       tmp_remote_dir[ln + 1] = 0;
	    }
	    dt |= (LEAVE_PENDING | LIST_ONLY | NOTFTP2HTML);
	    err = getftp(remote_host, remote_port, tmp_remote_dir,
			 LIST_FILENAME, ftp_acc, ftp_pass, &len, 0,
			 &res, &ftp_csock, &dt);
	    free(tmp_remote_dir);
	 }
	 tms = time_str();
	 switch (err)
	 {
	    case CONSOCKERR: case CONTIMEDOUT: case CONERROR:
	    case FTPSRVERR: case FTPUNKNOWNTYPE:
	       /* Various non-fatal errors produce continuation with
                  repeated logging-in. */
	       if (ftp_csock)
		  close(ftp_csock);
	       ftp_csock = 0;
	       if (local_name)
	       {
		  free(local_name);
		  local_name = NULL;
	       }
	       continue;
	       break;
	    case BINDERR: case CONPORTERR: case LISTENERR:
	    case FTPPORTERR: case ACCEPTERR:
	       /* Various non-fatal errors produce continuation. */
	       continue;
	       break;
	    case FTPLOGINC:
	       /* If login is incorrect, it can mean two things:
		  1. FTP server is too busy and does not allow logging
		  in, or
		  2. Login is incorrect.
		  If the number of tries is set to infinity, we must
		  take into account the possibility number two, and
		  set ntry to an arbitrary (presumably high, but not
		  infinite) value.  */
	       if (ntry == 0)
		  ntry = FTP_MAX_LINC;
	       close(ftp_csock);
	       ftp_csock = 0;
	       if (local_name)
	       {
		  free(local_name);
		  local_name = NULL;
	       }
	       continue;
	       break;
	    case HOSTERR: case CONREFUSED:
	    case FTPBADFILE:
	       /* Fatal errors -> return. */
	       if (ftp_csock)
		  close(ftp_csock);
	       unlink(files[where - 1]);
	       free(remote_host);
	       free(remote_dir);
	       if (local_name)
	       {
		  free(local_name);
		  local_name = NULL;
	       }
	       return err;
	       break;
	    case FTPNSFOD:
	       if (where)
	       {
		  err = RETRFINISHED; /* We'r lying to get past the
					 control. */
		  if (local_name)
		  {
		     free(local_name);
		     local_name = NULL;
		  }
		  continue;
	       }
	       else
	       {
		  if (ftp_csock)
		     close(ftp_csock);
		  free(remote_dir);
		  free(remote_host);
		  if (local_name)
		  {
		     free(local_name);
		     local_name = NULL;
		  }
		  return err;
	       }
	       break;
	    case FWRITEERR:
	       /* Another fatal. */
	       if (!quiet)
	       {
		  fprintf(lfile, "Can't write to \"%s\" (%s).\n",
			  where ? files[where - 1] : LIST_FILENAME,
			  mystrerror(errno));
	       }
	       /* I will not unlink *file on purpose. */
	       free(remote_host);
	       free(remote_dir);
	       if (ftp_csock)
		  close(ftp_csock);
	       if (local_name)
	       {
		  free(local_name);
		  local_name = NULL;
	       }
	       return err;
	       break;
	    case FTPRETRINT: case RETRFINISHED: case FTPRERR:
	       /* Deal with you later. */
	       break;
	    default:
	       /* All possibilities should have been exhausted. */
	       fprintf(lfile, "This should also not happen. Bailing out.\n");
	       exit(1);
	 }
	 if (err == RETRFINISHED)
	 {
	    if (!(dt & LIST_ONLY))
	    {
	       if (verbose)
	       {
		  fprintf(lfile, "%s URL [%ld] -> \"%s\" [%d]\n", tms, len,
			  where ? local_name : LIST_FILENAME, count);
	       }
	       else if (!quiet)
	       {
		  fprintf(lfile, "%s URL: %s [%ld] -> \"%s\" [%d]\n", tms,
			  url, len, where ? local_name : LIST_FILENAME,
			  count);
	       }
	    }
	    if (local_name)
	    {
	       free(local_name);
	       local_name = NULL;
	    }
	    continue;
	 }
	 else /* It must be FTPRETRINT or FTPRERR, which boil down to
		 the same thing. */
	 {
	    if (verbose && len != 0)
	    {
	       fprintf(lfile, "\nData connection closed at byte %ld. ", len);
	       printwhat(count, ntry);
	    }
	    if (local_name)
	    {
	       free(local_name);
	       local_name = NULL;
	    }
	    continue;
	 }
      } while ((!ntry || (count < ntry)) && err != RETRFINISHED);
      /* If we could not fetch the directory listing, we can do
         nothing. */
      if (!where && (ntry && count >= ntry))
      {
	 if (local_name)
	 {
	    free(local_name);
	    local_name = NULL;
	 }
	 break;
      }
   }
   if (files)
      for (i = 0; files[i]; i++)
	 free(files[i]);
   free(files);
   free(remote_host);
   free(remote_dir);
   if (ftp_csock)
      close(ftp_csock);
   return TRYLIMEXC;
}
