/* host.c
   A module for dealing with hosts */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#ifdef HAVE_SYS_SYSTEMINFO_H
#  include <sys/systeminfo.h>
#endif

#include "geturl.h"
#include "nmalloc.h"
#include "host.h"
#include "url.h"

extern int simple_check, quiet;
extern FILE *lfile;

#ifdef DEBUG
extern int debug;
#endif

/* A list of looked-up hosts */
hhost_t **hhash = NULL;

/* A NULL-terminated list of domain suffixes to
   follow */
char **domains = NULL;

/* The same as gethostbyname, but supports internet addresses of the
   form N.N.N.N */
struct hostent *
ngethostbyname(char *name)
{
   struct hostent *hp;
   unsigned long addr;

   if (inet_addr(name) != -1)
   {
      /* Yes, I know it is lame to call inet_addr twice, but I
         couldn't think of any other solution portable to 64-bit
         machines. */
      addr = (unsigned long)inet_addr(name);
      hp = gethostbyaddr((char *)&addr, sizeof (addr), AF_INET);
   }
   else
      hp = gethostbyname(name);
   return hp;
}

/* The routine used to determine the "real" name of the host, as
   viewed by the program. If host is referenced by more than one name,
   "real" name is considered to be the first one.
   
   If the host cannot be found in the hash table of already
   dealed-with hosts, try with its INET address. If this fails too,
   add it to the hash.  The routine does not call gethostbyname twice
   for the same host if it can possibly avoid it */
char *
realhost(char *host)
{
   int i, ind;
   struct in_addr in;
   struct hostent *hptr;
   char *inet_s;

#ifdef DEBUG
   if (debug)
      fprintf(lfile, "Checking for %s.\n", host);
#endif
   if (!hhash)
      hhash = create_hhash();
   ind = -1;
   for (i = 0; hhash[i]; i++)
      if (!strcasecmp(hhash[i]->hostname, host))
      {
	 ind = i;
	 if (hhash[i]->quality)
	    break;
      }
   if (hhash[i]) /* Found it with quality */
      return nstrdup(host);
   /* The fact that gethostbyname will get called makes
      it necessary to store it to the hash, to ensure
      that gethostbyname will not be called twice
      for the same string. However, the quality
      argument must be set appropriately.

      Note that add_hhash must be called *after* the
      realname search, or the quality would be
      always set to 0 */
   if (ind == -1) /* Not found, with or without quality */
   {
      hptr = ngethostbyname(host);
      if (!hptr)
	 return nstrdup(host);
      memcpy(&in.s_addr, *hptr->h_addr_list, sizeof(in.s_addr));
      inet_s = nstrdup(inet_ntoa(in));
   }
   else /* Found, without quality */
   {
      /* This case happens when host is on the list,
	 but not as first entry (the one with quality).
	 Then we just get its INET address and pick
	 up the first entry with quality. */
      inet_s = nstrdup(hhash[ind]->realname);
   }
   
   /* Now we certainly have the INET address. The following
      loop is guaranteed to pick either an entry with
      quality (because it is the first), or nothing */
   for (i = 0; hhash[i]; i++)
      if (!strcasecmp(hhash[i]->realname, inet_s))
	 break;
   if (hhash[i]) /* Found in the hash list, as realname */
   {
      /* Add_hhash will set quality to 0 */
      add_hhash(&hhash, host, inet_s);
      free(inet_s);
      return nstrdup(hhash[i]->hostname);
   }
   /* Since this is really the first time this host is
      encountered, set quality to 1 */
   add_hhash(&hhash, host, inet_s);
   hhash[i]->quality = 1;
   free(inet_s);
   return nstrdup(host);
}

/* This routine compares two hostnames (out of URL-s), taking care of
   aliases.  It uses realhost to determine a unique hostname for each
   of two hosts. If simple_check is set, only strcmp is called. */
int
same_host(char *u1, char *u2)
{
   char *s1, *s2, *real1, *real2;
   int i;

   /* Assumed HTTP. */
   if (!strncmp(u1, "http://", 7))
      u1 += 7;
   if (!strncmp(u2, "http://", 7))
      u2 += 7;
   s1 = nmalloc(1);
   *s1 = 0;
   for (i = 0; *u1 && *u1 != '/' && *u1 != ':'; u1++, i++)
   {
      s1 = nrealloc(s1, i + 2);
      s1[i] = *u1;
      s1[i + 1] = 0;
   }
   s2 = nmalloc(1);
   *s2 = 0;
   for (i = 0; *u2 && *u2 != '/' && *u2 != ':'; u2++, i++)
   {
      s2 = nrealloc(s2, i + 2);
      s2[i] = *u2;
      s2[i + 1] = 0;
   }
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "Comparing hosts %s and %s\n", s1, s2);
#endif
   if (!strcmp(s1, s2))
   {
      free(s1);
      free(s2);
      return 1;
   }
   else if (simple_check)
      return 0;
   real1 = realhost(s1);
   real2 = realhost(s2);
   return strcasecmp(real1, real2) ? 0 : 1;
}

/* Create hash-list of hhost_t host representations */
hhost_t **
create_hhash(void)
{
   hhost_t **res;
   res = nmalloc(sizeof(hhost_t *));
   *res = NULL;
   return res;
}

/* Add an entry to host list */
void
add_hhash(hhost_t ***hhash, char *s1, char *s2)
{
   int i;

#ifdef DEBUG
   if (debug)
      fprintf(lfile, "Adding %s to hash (realname %s)\n", s1, s2);
#endif
   for (i = 0; (*hhash)[i]; i++);
   *hhash = nrealloc(*hhash, sizeof(hhost_t *) * (i + 2));
   (*hhash)[i] = nmalloc(sizeof(hhost_t));
   (*hhash)[i]->hostname = nstrdup(s1);
   (*hhash)[i]->realname = nstrdup(s2);
   (*hhash)[i]->quality = 0;
   (*hhash)[i + 1] = NULL;
}

/* This routine frees memory allocated by host hash table, and resets
   it to NULL */
void
free_hhash(hhost_t **hhash)
{
   int i;

   if (!hhash)
      return;
   for (i = 0; hhash[i]; i++)
   {
      free(hhash[i]->hostname);
      free(hhash[i]->realname);
      free(hhash[i]);
   }
   free(hhash);
   hhash = NULL;
}

/* Determine whether a URL is acceptable to be followed,
   according to a list of domains to accept */
int
accept_domain(char *constr)
{
   char *host, *dir;
   int port, res;

   if (!domains)
      return 1;
   /* Standard allocation procedures before calling parseurl */
   host = nmalloc(strlen(constr) + 1);
   dir = nmalloc(strlen(constr) + 1);
   parseurl(constr, host, &port, dir);
   free(dir);
   res = sufmatch(domains, host);
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "I do%s think %s will be acceptable...\n",
	      res ? "" : " not", host);
#endif
   free(host);
   return res;
}

/* Check whether a host is matched in the list of domains */
int
sufmatch(char **list, char *what)
{
   int i, j, k;

   for (i = 0; list[i]; i++)
   {
      for (j = strlen(list[i]), k = strlen(what); j >= 0 && k >= 0; j--, k--)
	 if (list[i][j] != what[k])
	    break;
      /* The domain must be first to reach to beginning. */
      if (j == -1)
	 return 1;
   }
   return 0;
}

/* Returns the adress of a static array of four unsigned chars where
   the INET address of local host is. */
unsigned char *
inet_address(void)
{
   static char hostname[256];
   struct hostent *hptr;
   static unsigned char addr[4];
   static int first = 1;

   /* Do it only the first time. */
   if (first)
   {
      addr[0] = 0; addr[1] = 0; addr[2] = 0; addr[3] = 0;
#ifdef MY_HOST
      strcpy(hostname, MY_HOST);
#elif defined(HAVE_GETHOSTNAME)
      if (gethostname(hostname, 255) < 0)
	 return NULL;
#else /* you'd better have sysinfo */
      if (sysinfo(SI_HOSTNAME, hostname, 255) < 0)
	 return NULL;
#endif /* MY_HOST & co. */
      hptr = ngethostbyname(hostname);
      if (!hptr || !hptr->h_addr_list || !*hptr->h_addr_list)
	 return NULL;
      strncpy((char *)addr, *hptr->h_addr_list, 4);
      first = 0;
   }
   return addr;
}

/* Get email address of the form username@host.domain suitable for
   anonymous FTP passwords. If you have problems, hard-code your
   hostname and domainname by defining MY_HOST and MY_DOMAIN in
   config.h. If none of the available methods of getting host name and
   domain name works, returns username@. If it cannot get username,
   the program quits, and such Unix is not worthy of mentioning. */
char *
ftp_getaddress(void)
{
   static char address[256];
   int i, pos;
   static int first = 1;

   /* Do it only the first time, since it won't change. */
   if (first)
   {
      first = 0;
      if (!cuserid(address))
      {
	 if (!quiet)
	    fprintf(lfile, "Cannot find your login name.\n");
	 exit(1);
      }
      i = strlen(address);
      address[i++] = '@';
      pos = i;
      /* pos holds the position after '@'. */
#ifdef MY_HOST
      strcpy(address + i, MY_HOST);
#elif defined(HAVE_GETHOSTNAME)
      if (gethostname(address + i, 256 - i - 1) < 0)
	 return address;
#else /* you'd better have sysinfo */
      if (sysinfo(SI_HOSTNAME, address + i, 256 - i - 1) < 0)
	 return address;
#endif
      i = strlen(address);
#ifdef MY_DOMAIN
      address[i++] = '.';
      strcat(address, MY_DOMAIN);
#elif defined(HAVE_GETDOMAINNAME)
      address[i++] = '.';
      if (getdomainname(address + i, 256 - i - 1) < 0)
      {
	 address[pos] = 0;
	 return address;
      }
#else /* you'd better have sysinfo */
      address[i++] = '.';
      if (sysinfo(SI_SRPC_DOMAIN, address + i, 256 - i - 1) < 0)
      {
	 address[pos] = 0;
	 return address;
      }
#endif
   }
   return address;
}

