/* http.c
   HTTP support module */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <sys/stat.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <ctype.h>

#include "geturl.h"
#include "nmalloc.h"
#include "http.h"
#include "connect.h"
#include "loop.h"
#include "url.h"

extern int verbose, quiet, follow_location;

extern FILE *lfile;
extern char *version_string;

#ifdef DEBUG
extern int debug;
#endif

extern int errno;

/* Function to fetch a header from socket/file descriptor fd. Most
   errors are handled.

   The trailing CRLF or LF are stripped from the header, and it is
   zero-terminated. */
uerr_t
fetch_next_header(int fd, char *hdr, int maxlen)
{
   int i, res;

   for (i = 0; i < maxlen; i++)
   {
      res = iread(fd, hdr + i, 1);
      if (res == 1)
      {
	 if (hdr[i] == '\n')
	 {
	    hdr[i] = 0;
	    /* Get rid of '\r'. */
	    if (i > 0 && hdr[i - 1] == '\r')
	       hdr[i - 1] = 0; /* He he */
	    break;
	 }
      }
      else if (res == 0)
	 return HEOF;
      else
	 return HERR;
   }
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "%s\n", hdr);
#endif
   if (hdr[i])
      return HLEXC;
   else
      return HOK;
}

/* Return the content length of the document body, if this is
   Content-length header, -1 otherwise. */
long
hgetlen(char *hdr)
{
   long len;
   
   if (strncasecmp(hdr, CONTLEN_H, strlen(CONTLEN_H)))
      return -1;
   len = 0;
   for (hdr += strlen(CONTLEN_H); isdigit(*hdr); hdr++)
   {
      len *= 10;
      len += *hdr - '0';
   }
   return len;
}

/* Sucks the contents of file descriptor fd, until it is closed, or a
   read error occurs. The data is read in parts BUFFER_SIZE bytes
   long, and stored to stream fp, which should have been open for
   writing.

   If verbose is set, the progress is also shown. Variable restval is
   used only to reset the progress value to current restval instead to
   0, and will be ignored if set to 0.

   The function exits and returns codes of 0, 1 and 2 if the
   connection was closed, there was a read error, or if it could not
   write to the output stream, respectively. */
int
get_contents(int fd, FILE *fp, long *len, long restval)
{
   int res;
   static char c[BUFFER_SIZE]; /* Static, to avoid stack overload */

   *len = restval;
   if (restval)
      show_progress(0, restval);
   do
   {
      res = iread(fd, c, BUFFER_SIZE);
      if (res > 0)
      {
	 if (!fwrite(c, sizeof(char), res, fp))
	    return -2;
	 if (verbose)
	    show_progress(*len, res);
	 *len += res;
      }
   } while (res > 0);
   if (res < -1)
      res = -1;
   if (verbose)
      fprintf(lfile, "\n\n");
   return res;
}

/* Returns the malloc-ed copy of the type of the header hdr, to the
   first ';', or NULL if the header does not begin with CONTTYPE_H
   string. */
char *
hgettype(char *hdr)
{
   char *type;

   if (strncasecmp(hdr, CONTTYPE_H, strlen(CONTTYPE_H)))
      return NULL;
   for (type = hdr + strlen(CONTTYPE_H); *type && *type != ';'; type++);
   *type = 0;
   type = nstrdup(hdr + strlen(CONTTYPE_H));
   return type;
}

/* Returns the malloc-ed copy of the location of the document, if the
   string hdr begins with LOCATION_H, or NULL. */
char *
hgetlocation(char *hdr)
{
   char *location;

   if (strncmp(hdr, LOCATION_H, strlen(LOCATION_H)))
      return NULL;
   location = nstrdup(hdr + strlen(LOCATION_H));
   return location;
}

/* Retrieves a document through HTTP protocol, with the parameters
   named in the argument list. If the location header has changed and
   follow_location is set, it returns NEWLOCATION. It closes the
   network socket (although it already gets closed by the server). If
   it receives an error from the functions below it, it will print it
   if there is enough information to do so (almost always). It will
   always return the error to the upper function
   (i.e. retrieve_url()). */
uerr_t
gethttp(char *remote_host, int remote_port, char *remote_dir,
	FILE *fp, long *l1, long *l2, int *res, char **loc,
	int *doctype, char *error)
{
   char *request, *type, *location;
   static char hdr[MAXHLEN];
   int sock, hcount;
   long contlen;
   uerr_t err;

   *loc = NULL;
   *error = 0;
   /* First, let's make the request. */
   request = nmalloc(strlen(remote_dir) + strlen(version_string) + 50);
   sprintf(request,
	   "GET %s HTTP/1.0\r\nUser-Agent: %s\r\nAccept: */*\r\n\r\n",
	   remote_dir, version_string);
   /* Then, let's establish the connection. */
   if (verbose)
      fprintf(lfile, "Connecting to %s:%d... ", remote_host, remote_port);
   err = make_connection(&sock, remote_host, remote_port);
   switch (err)
   {
      case HOSTERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "No such host %s.\n", remote_host);
	 }
	 free(request);
	 return HOSTERR;
	 break;
      case CONSOCKERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Socket error (%s).\n", mystrerror(errno));
	 }
	 free(request);
	 return CONSOCKERR;
	 break;
      case CONTIMEDOUT:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Connection to %s timed out.\n", remote_host);
	 }
	 free(request);
	 close(sock);
	 return CONTIMEDOUT;
	 break;
      case CONREFUSED:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Connection to %s refused.\n", remote_host);
	 }
	 free(request);
	 close(sock);
	 return CONREFUSED;
      case CONERROR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "connect: %s\n", mystrerror(errno));
	 }
	 free(request);
	 close(sock);
	 return CONERROR;
	 break;
      case NOCONERROR:
	 if (verbose)
	    fprintf(lfile, "Connected!\n");
	 /* Everything is fine! */
	 break;
      default:
	 fprintf(lfile, "This condition should not have been reached.\n");
	 exit(1);
	 break;
   }
   
   /* Send the request to server */
   write(sock, request, strlen(request));
   if (verbose)
      fprintf(lfile, "HTTP request sent, fetching headers... ");
   free(request);
   contlen = -1;
   type = location = NULL;
   *doctype &= ~RETROKF;
   hcount = 0;
   do
   {
      ++hcount;
      err = fetch_next_header(sock, hdr, MAXHLEN);
      if (verbose)
         fprintf(lfile, "%d ", hcount);
      if (err == HEOF)
      {
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "End of file while parsing headers.\n");
	 }
	 if (type)
	    free(type);
	 if (location)
	    free(location);
	 return HEOF;
      }
      else if (err == HERR)
      {
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Read error (%s) in headers.\n",
	            errno ? mystrerror(errno) : "timeout");
	 }
	 if (type)
	    free(type);
	 if (location)
	    free(location);
	 return HERR;
      }
      if (hcount == 1)
      {
	 int i = 0;
	 
	 if (!strncasecmp(hdr, HTTP200_H, strlen(HTTP200_H)))
	    *doctype |= RETROKF;
	 while (hdr[i] && hdr[i] != ' ')
	    ++i;
	 ++i;
	 strcpy(error, hdr + i);
      }
      if (contlen == -1)
	 contlen = hgetlen(hdr);
      if (!type)
	 type = hgettype(hdr);
      if (!location)
	 location = hgetlocation(hdr);
   } while (strlen(hdr));
   if (type)
      (!strncasecmp(type, TEXTHTML_S, strlen(TEXTHTML_S))) ?
	 (*doctype |= TEXTHTML) : (*doctype &= ~TEXTHTML);
   else
      *doctype &= ~TEXTHTML; /* NOT text/html by default */
   *l2 = contlen;
   if (verbose)
   {
      fprintf(lfile, "done!\n");
      if (location)
      {
	 fprintf(lfile, "Location: %s", location);
	 if (follow_location)
	    fprintf(lfile, " [following]");
	 fprintf(lfile, "\n");
      }
      if (!(location && follow_location))
      {
	 fprintf(lfile, "Length: ");
	 if (contlen != -1)
	    fprintf(lfile, "%s", legible(contlen));
	 else
	    fprintf(lfile, "unspecified");
	 if (type)
	    fprintf(lfile, " [%s]\n", type);
	 else
	    fprintf(lfile, "\n");
      }
   }
   if (location && follow_location)
   {
      close(sock);
      if (type)
	 free(type);
      *loc = location;
      return NEWLOCATION;
   }
   *res = get_contents(sock, fp, l1, 0);
   close(sock);
   if (type)
      free(type);
   if (location)
      free(location);
   if (*res == -2)
      return FWRITEERR;
   return RETRFINISHED;
}

/* Retrieves the document similarly to gethttp, but through a proxy
   gateway. The routine is almost a copy of gethttp, and behaves
   in a similar fashion. */
uerr_t
proxy_retrieve(char *proxy, char *url, FILE *fp, long *l1, long *l2,
	       int *res, char **loc, int *doctype, char *error)
{
   char *request, *type, *location, *dum, *host;
   static char hdr[MAXHLEN];
   int sock, hcount, port;
   long contlen;
   uerr_t err;

   host = nmalloc(strlen(proxy));
   dum = nmalloc(strlen(proxy));
   err = parseurl(proxy, host, &port, dum);
   free(dum);
   if (err != URLHTTP)
   {
      free(host);
      if (!quiet)
	 fprintf(lfile, "Error in proxy address %s.\n", proxy);
      return PROXERR;
   }
   *loc = NULL;
   /* First, let's make the request. */
   request = nmalloc(strlen(url) + strlen(version_string) + 50);
   /* Should I send <CR> also? */
   sprintf(request,
	   "GET %s HTTP/1.0\nUser-Agent: %s\nAccept: */*\n\n",
	   url, version_string);
   /* Then, let's establish the connection. */
   if (verbose)
      fprintf(lfile, "Connecting to %s:%d... ", host, port);
   err = make_connection(&sock, host, port);
   switch (err)
   {
      case HOSTERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "No such host %s.\n", host);
	 }
	 free(request);
	 free(host);
	 return HOSTERR;
	 break;
      case CONSOCKERR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Socket error (%s).\n", mystrerror(errno));
	 }
	 free(request);
	 free(host);
	 return CONSOCKERR;
	 break;
      case CONTIMEDOUT:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Connection to %s timed out.\n", host);
	 }
	 free(request);
	 close(sock);
	 free(host);
	 return CONTIMEDOUT;
	 break;
      case CONREFUSED:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Connection to %s refused.\n", host);
	 }
	 free(request);
	 close(sock);
	 free(host);
	 return CONREFUSED;
      case CONERROR:
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "connect: %s\n", mystrerror(errno));
	 }
	 free(request);
	 close(sock);
	 free(host);
	 return CONERROR;
	 break;
      case NOCONERROR:
	 if (verbose)
	    fprintf(lfile, "Connected!\n");
	 /* Everything is fine! */
	 break;
      default:
	 fprintf(lfile, "This condition should not have been reached.\n");
	 exit(1);
	 break;
   }
   
   /* Send the request to server */
   write(sock, request, strlen(request));
   if (verbose)
      fprintf(lfile, "Proxy request sent, fetching headers... ");
   /* Safe to free now. */
   free(request);
   free(host);
   
   contlen = -1;
   type = location = NULL;
   *doctype &= ~RETROKF;
   hcount = 0;
   do
   {
      ++hcount;
      err = fetch_next_header(sock, hdr, MAXHLEN);
      if (err == HEOF)
      {
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "End of file while parsing headers.\n");
	 }
	 if (type)
	    free(type);
	 if (location)
	    free(location);
	 return HEOF;
      }
      else if (err == HERR)
      {
	 if (!quiet)
	 {
	    if (verbose)
	       fprintf(lfile, "\n");
	    fprintf(lfile, "Read error (%s) in headers.\n",
	            errno ? mystrerror(errno) : "timeout");
	 }
	 if (type)
	    free(type);
	 if (location)
	    free(location);
	 return HERR;
      }
      if (hcount == 1)
      {
	 int i = 0;
	 if (!strncasecmp(hdr, HTTP200_H, strlen(HTTP200_H)))
	    *doctype |= RETROKF;
	 while (hdr[i] && hdr[i] != ' ')
	    ++i;
	 ++i;
	 strcpy(error, hdr + i);
      }
      if (contlen == -1)
	 contlen = hgetlen(hdr);
      if (!type)
	 type = hgettype(hdr);
      if (!location)
	 location = hgetlocation(hdr);
   } while (strlen(hdr));
   if (type)
      (!strncasecmp(type, TEXTHTML_S, strlen(TEXTHTML_S))) ?
	 (*doctype |= TEXTHTML) : (*doctype &= ~TEXTHTML);
   else
      *doctype &= ~TEXTHTML; /* NOT text/html by default */
   *l2 = contlen;
   if (verbose)
   {
      fprintf(lfile, "done!\n");
      if (location)
      {
	 fprintf(lfile, "Location: %s", location);
	 if (follow_location)
	    fprintf(lfile, " [following]");
	 fprintf(lfile, "\n");
      }
      if (!(location && follow_location))
      {
	 fprintf(lfile, "Length: ");
	 if (contlen != -1)
	    fprintf(lfile, "%s", legible(contlen));
	 else
	    fprintf(lfile, "unspecified");
	 if (type)
	    fprintf(lfile, " [%s]\n", type);
	 else
	    fprintf(lfile, "\n");
      }
   }
   if (location && follow_location)
   {
      close(sock);
      if (type)
	 free(type);
      *loc = location;
      return NEWLOCATION;
   }
   *res = get_contents(sock, fp, l1, 0);
   close(sock);
   if (type)
      free(type);
   if (location)
      free(location);
   if (*res == -2)
      return FWRITEERR;
   return RETRFINISHED;
}

/* Legible -- return a string with the legible printed number. Not
   very nice, using sprintf(), but it works fine. */
char *
legible(long l)
{
   static char buf[20];
   char inbuf[20];
   int i, i1, mod;
   char *ptr, *in;

   /* Fill the buffer. */
   sprintf(inbuf, "%ld", l);
   /* Reset the pointers. */
   ptr = buf;
   in = inbuf;
   /* If the number is negative, shift the pointers. */
   if (*in == '-')
   {
      *ptr++ = '-';
      ++in;
   }
   /* How many digits before the first separator? */
   mod = strlen(in) % 3;
   /* Insert them. */
   for (i = 0; i < mod; i++)
      *ptr++ = in[i];
   /* Now insert the rest of them, putting separator before every
      third digit. */
   for (i1 = i, i = 0; in[i1]; i++, i1++)
   {
      if (i % 3 == 0 && i1 != 0)
	 *ptr++ = LEGIBLE_SEPARATOR;
      *ptr++ = in[i1];
   }
   /* Insert zero at the end. */
   *ptr = 0;
   return buf;
}

