/* init.c
   Module for reading/parsing the initialization file */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#ifdef HAVE_PWD
#  include <pwd.h>
#endif
#include <errno.h>

#include "geturl.h"
#include "nmalloc.h"
#include "init.h"
#include "url.h"
#include "ftp.h"
#include "host.h"

/* AAAaaaaaaaaaargh... global variables! */
int verbose;         /* Are we verbose? */
int quiet;           /* Are we quiet? */
int ntry;            /* Number of tries per URL */
int kill_error;      /* Do we kill or save error documents? */
int kill_longer;     /* Do we reject messages with *more* data than
			specified in content-length? */
int follow_location; /* Do we follow Location: header? */
int recursive;       /* Are we recursive? */
int spanhost;        /* Do we span across hosts in recursion? */
int relative_only;   /* Follow only relative links. */
int simple_check;    /* Should we use simple checking (strcmp) or do
			we create a host hash and call gethostbyname? */
int maxreclevel;     /* Maximum level of recursion */
int prefix_files;    /* Do we give files a prefix, or do we construct
			their name from the URL? */
int dirstruct;       /* Do we build the directory structure as we go
			along? */
int dirmode;         /* Permissions for direcories in directory
			structure */
int add_hostdir;     /* Do we add hostname directory? */
int noclobber;       /* Disables clbbering of existing data. */
char *dir_prefix;    /* The top of directory tree */
char *fileprefix;    /* Prefix of files */
char *lfilename;     /* Log filename */
char *input_filename;/* Input filename */
int force_html;      /* Is the input file an HTML file? */

char **accepts;      /* List of extensions to accept */
char **rejects;      /* List of extensions to reject */

extern char **domains; /* See host.c */

int follow_ftp;      /* Are FTP URL-s followed in recursive retrieving? */
int always_rest;     /* Always use REST. */
char *ftp_acc;       /* FTP username */
char *ftp_pass;      /* FTP password */
int ftp_glob;        /* FTP globbing */

int use_proxy;       /* Do we use proxy? */
char *http_proxy, *ftp_proxy, *no_proxy;
char *base_href;
int timeout;         /* The value of read timeout in seconds. */
int use_robots;      /* Do we heed robots.txt? */

extern FILE *lfile;         /* Global log file */
#ifdef DEBUG
int debug;       /* Debugging on/off */
#endif /* DEBUG */

extern hhost_t **hhash;

extern int errno;

/* List of recognized commands, each consisting of a string-name and
   ID. */
struct cmd commands[] = {
   { "accept", ACCEPT },
   { "addhostdir", ADDHOSTDIR },
   { "alwaysrest", ALWAYSREST },
   { "basehref", BASEHREF },
   { "debug", DEBUG_ },
   { "dirmode", DIRMODE },
   { "dirprefix", DIRPREFIX },
   { "dirstruct", DIRSTRUCT },
   { "domains", DOMAINS },
   { "fileprefix", FILEPREFIX },
   { "followftp", FOLLOWFTP },
   { "forcehtml", FORCEHTML },
   { "ftpproxy", FTPPROXY },
   { "glob", GLOB },
   { "httpproxy", HTTPPROXY },
   { "input", INPUT },
   { "killerror", KILLERROR },
   { "killlonger", KILLLONGER },
   { "logfile", LOGFILE },
   { "login", LOGIN },
   { "noclobber", NOCLOBBER },
   { "nolocation", NOLOCATION },
   { "noproxy", NOPROXY },
   { "numtries", NUMTRIES },
   { "passwd", PASSWD },
   { "prefixfiles", PREFIXFILES },
   { "quiet", QUIET },
   { "reclevel", RECLEVEL },
   { "recursive", RECURSIVE },
   { "reject", REJECT },
   { "relativeonly", RELATIVEONLY },
   { "robots", ROBOTS },
   { "simplehostcheck", SIMPLEHOSTCHECK },
   { "spanhosts", SPANHOSTS },
   { "timeout", TIMEOUT },
   { "useproxy", USEPROXY },
   { "verbose", VERBOSE },
   { NULL, LAST }
};


/* Return command ID of a valid command, -1 if none. */
cmdid
comind(struct cmd *commands, char *com)
{
   int i;

   for (i = 0; commands[i].name; i++)
      if (!strcasecmp(commands[i].name, com))
	 return commands[i].id;
   return NONE;
}

/* Reset the variables to default values. */
void
defaults(void)
{
#ifdef DEBUG
   debug = 0;
#endif
   verbose = -1;
   quiet = 0;
   
   kill_error = 1;
   kill_longer = 0;
   lfilename = NULL;
   input_filename = NULL;
   fileprefix = nstrdup(DEFAULT_FILE_PREFIX);
   ntry = DEFAULT_NUMBER_OF_TRIES;
   recursive = 0;
   follow_location = 1;
   maxreclevel = DEFAULT_MAXRECLEVEL;
   spanhost = 0;
   relative_only = 0;
   accepts = rejects = NULL;
   prefix_files = 0;
   dirstruct = 1;
   simple_check = 0;
   dirmode = DEFAULT_DIRMODE;
   add_hostdir = 1;
   dir_prefix = nstrdup(DEFAULT_DIRPREFIX);
   noclobber = 0;
   
   follow_ftp = 0;
   always_rest = 0;
   ftp_acc = nstrdup(DEFAULT_FTP_ACCT);
   ftp_pass = nstrdup(ftp_getaddress());
   ftp_glob = 1;

   use_proxy = 1;
   http_proxy = ftp_proxy = NULL;
   if ((no_proxy = getenv("no_proxy")))
      no_proxy = nstrdup(no_proxy);
   base_href = NULL;

   timeout = 300;
   use_robots = 1;
}

/* Return the path of .geturlrc. */
char *
init_path(void)
{
   char *env, *home, *path;
   struct stat buf;
   int err;
#ifdef HAVE_PWD
   struct passwd *pwd;
#endif

   /* Try the environment. */
   if ((env = getenv("GETURLRC")))
   {
      path = nstrdup(env);
      err = stat(path, &buf);
      if (err == 0)
	 return path;
      free(path);
      fprintf(stderr, "Cannot load .geturlrc.\n");
      exit(1);
   }
   /* If that failed, try ./.geturlrc. */
   path = nstrdup(".geturlrc");
   err = stat(path, &buf);
   if (err == 0)
      return path;
   free(path);
   /* If that failed too, try $HOME/.geturlrc or $LOGDIR/.geturlrc. */
   home = getenv("HOME");
   if (!home)
      home = getenv("LOGDIR");
#ifdef HAVE_PWD
   if (!home)
   {
      pwd = getpwuid(getpid());
      if (!pwd || !pwd->pw_dir)
	 return NULL;
      home = pwd->pw_dir;
   }
#endif
   if (!home)
      return NULL;
   path = nmalloc(strlen(home) + 1 + strlen(".geturlrc") + 1);
   sprintf(path, "%s/.geturlrc", home);
   err = stat(path, &buf);
   if (err == -1)
   {
      free(path);
      return NULL;
   }
   return path;
}

/* Initialize variables from .geturlrc. */
void
initialize(void)
{
   char *path, *com, *val;
   FILE *fp;
   static char line[INITLINESIZE];
   int stat, ln;

   /* Load defaults. */
   defaults();
   if (!(path = init_path()))
      return;
   fp = fopen(path, "r");
   if (!fp)
   {
      fprintf(stderr, "Cannot read %s (%s).\n", path, mystrerror(errno));
      return;
   }
   /* Reset line number. */
   ln = 1;
   /* While there are lines in the file: */
   while (fgets(line, INITLINESIZE, fp))
   {
      /* Parse the line. */
      stat = parse_line(line, &com, &val);
      /* If everything is OK, set the value. */
      if (stat == 1)
      {
	 if (!setval(com, val))
	    fprintf(stderr, "Syntax error in line %d.\n", ln);
	 free(com);
	 free(val);
      }
      else if (stat == 0)
	 fprintf(stderr, "Syntax error in line %d.\n", ln);
      ++ln;
   }
   free(path);
}

/* Parse the line pointed by line, with the syntax:
   <sp>* command <sp>* = <sp>* value <newline>
   Uses malloc to allocate space for command and value.
   If the line is invalid, data is freed and 0 is returned.

   Return values:
    1 - success
    0 - failure
   -1 - comment */
int
parse_line(char *line, char **com, char **val)
{
   int i, j;

   /* Skip spaces. */
   for (i = 0; line[i] == ' ' || line[i] == '\t'; i++);

   /* Skip empty or hashed lines. */
   if (!line[i] || line[i] == '\n' || line[i] == '#')
      return -1;
   
   /* Allocate command. */
   *com = nmalloc(1);
   (*com)[0] = 0;
   for (j = 0; ((line[i] > 'A' && line[i] < 'z') ||
		line[i] == '_' || line[i] == '-'); i++)
   {
      if (line[i] == '_' || line[i] == '-')
	 continue;
      *com = nrealloc(*com, j + 2);
      (*com)[j] = tolower(line[i]);
      (*com)[j + 1] = 0;
      ++j;
   }
   /* The next char should be space or '='. */
   if (!strchr(" \t\n=", line[i]))
   {
      free(*com);
      return 0;
   }
   /* If the command is invalid, exit now. */
   if (comind(commands, *com) == NONE)
   {
      free(*com);
      return 0;
   }
   /* Skip spaces before '='. */
   for (; strchr("\t\n ", line[i]); i++);
   /* If '=' not found, bail out. */
   if (line[i] != '=')
      return 0;
   /* Skip spaces after '='. */
   for (++i; strchr("\t\n ", line[i]); i++);
   /* Get the ending position. */
   for (j = i; line[i] && line[i] != '\n'; i++);
   /* Allocate *val, and copy from line. */
   (*val) = nmalloc(i - j + 1);
   strncpy(*val, line + j, i - j);
   (*val)[i - j] = 0;
   return 1;
}

/* Do the job for each line.  No fatals -- error signal prints a
   warning and resets to default value. */
int
setval(char *com, char *val)
{
   int flag;
   cmdid cmd;

   if (!com || !val)
      return 0;
   cmd = comind(commands, com);
   if (cmd == NONE)
   {
#ifdef DEBUG
      fprintf(stderr, "Wrong var %s, value %s.\n", com, val);
#endif
      return 0;
   }
   switch (cmd)
   {
      case ACCEPT:
	 accepts = sepstring(val);
	 break;
      case ADDHOSTDIR:
	 return setonoff(&add_hostdir, com, val);
	 break;
      case ALWAYSREST:
	 return setonoff(&always_rest, com, val);
      case BASEHREF:
	 if (base_href)
	    free(base_href);
	 base_href = nstrdup(val);
	 break;
      case DEBUG_:
#ifdef DEBUG
	 return setonoff(&debug, com, val);
#else
	 fprintf(stderr, "%s: debug support not loaded.\n", com);
	 return 0;
#endif
	 break;
      case DIRMODE:
	 flag = getmode(val);
	 if (flag == -1)
	 {
	    fprintf(stderr, "%s: Invalid mode specification.\n", val);
	    return 0;
	 }
	 dirmode = DEFAULT_DIRMODE;
	 break;
      case DIRPREFIX:
	 free(dir_prefix);
	 dir_prefix = nstrdup(val);
	 break;
      case DIRSTRUCT:
	 return setonoff(&dirstruct, com, val);
	 break;
      case DOMAINS:
	 domains = sepstring(val);
	 break;
      case FILEPREFIX:
	 free(fileprefix);
	 fileprefix = nstrdup(val);
	 break;
      case FOLLOWFTP:
	 return setonoff(&follow_ftp, com, val);
	 break;
      case FORCEHTML:
	 return setonoff(&force_html, com, val);
	 break;
      case FTPPROXY:
	 if (ftp_proxy)
	    free(ftp_proxy);
	 ftp_proxy = nstrdup(val);
	 break;
      case GLOB:
	 return setonoff(&ftp_glob, com, val);
	 break;
      case HTTPPROXY:
	 if (http_proxy)
	    free(http_proxy);
	 http_proxy = nstrdup(val);
	 break;
      case INPUT:
	 input_filename = nstrdup(val);
	 break;
      case KILLERROR:
	 return setonoff(&kill_error, com, val);
	 break;
      case KILLLONGER:
	 return setonoff(&kill_longer, com, val);
	 break;
      case LOGFILE:
	 if (lfilename)
	    free(lfilename);
	 lfilename = nstrdup(val);
	 break;
      case LOGIN:
	 free(ftp_acc);
	 ftp_acc = nstrdup(val);
	 break;
      case NOCLOBBER:
	 return setonoff(&noclobber, com, val);
	 break;
      case NOLOCATION:
	 flag = setonoff(&follow_location, com, val);
	 follow_location = !follow_location;
	 return flag;
	 break;
      case NOPROXY:
	 if (no_proxy)
	    free(no_proxy);
	 no_proxy = nstrdup(val);
	 break;
      case NUMTRIES:
	 return setnuminf(&ntry, com, val, DEFAULT_NUMBER_OF_TRIES);
	 break;
      case PASSWD:
	 free(ftp_pass);
	 ftp_pass = nstrdup(val);
	 break;
      case PREFIXFILES:
	 return setonoff(&prefix_files, com, val);
	 break;
      case QUIET:
	 return setonoff(&quiet, com, val);
	 break;
      case RECLEVEL:
	 return setnuminf(&maxreclevel, com, val, DEFAULT_MAXRECLEVEL);
	 break;
      case RECURSIVE:
#ifndef NO_RECURSION
	 return setonoff(&recursive, com, val);
#else
	 fprintf(stderr, "Recursion is not enabled.\n");
#endif
	 break;
      case REJECT:
	 rejects = sepstring(val);
	 break;
      case RELATIVEONLY:
	 return setonoff(&relative_only, com, val);
	 break;
      case ROBOTS:
	 return setonoff(&use_robots, com, val);
	 break;
      case SIMPLEHOSTCHECK:
	 return setonoff(&simple_check, com, val);
	 break;
      case SPANHOSTS:
	 return setonoff(&spanhost, com, val);
	 break;
      case TIMEOUT:
	 return setnuminf(&timeout, com, val, DEFAULT_TIMEOUT);
	 break;
      case USEPROXY:
	 return setonoff(&use_proxy, com, val);
	 break;
      case VERBOSE:
	 return setonoff(&verbose, com, val);
	 break;
      case NONE: case LAST: default:
	 printf("Error condition, bailing out.\n");
	 exit(1);
   }
   return 1;
}

/* Return 0 if off, 1 if on, -1 if error. */
int
onoff(char *val)
{
   if (!strcasecmp(val, "on"))
      return 1;
   else if (!strcasecmp(val, "off"))
      return 0;
   else
      return -1;
}

/* Set the integer value pointed to by what on value represented in
   string val. String com is used only for error messages. */
int
setonoff(int *what, char *com, char *val)
{
   int flag;

   flag = onoff(val);
   if (flag == -1)
   {
      fprintf(stderr, "%s: Please specify on or off.\n", com);
      return 0;
   }
   *what = flag;
   return 1;
}

/* Similar to setonoff, but with "inf" meaning zero, and having e
   default value for invalid numbers. */
int
setnuminf(int *what, char *com, char *val, int defl)
{
   if (!strcasecmp(val, "inf"))
      *what = 0;
   else if ((*what = myatoi(val)) == -1)
   {
      fprintf(stderr, "%s: Invalid specification.\n", val);
      *what = defl;
      return 0;
   }
   return 1;
}

/* Return the int value of a positive integer written in a string,
   or -1 if an error was encountered */
int
myatoi(char *s)
{
   int res;
   char *orig;

   orig = s;
   for (res = 0; *s && isdigit(*s); s++)
   {
      res *= 10;
      res += *s - '0';
   }
   if (*s || orig == s)
      return -1;
   else
      return res;
}

/* Return the decimal representation of octal mode three-digit
   permission string, or -1 if incorrect. */
int
getmode(char *s)
{
   if (strlen(s) != 3 || !isdigit(s[0]) || !isdigit(s[1]) || !isdigit(s[2]))
      return -1;
   return (s[2] - '0') + 8 * (s[1] - '0') + 64 * (s[0] - '0');
}

/* Free the memory allocated by global variables. */
void
freemem(void)
{
   if (lfilename)
      free(lfilename);
   free(fileprefix);
   if (input_filename)
      free(input_filename);
   freelists();
   free_hhash(hhash);
   free(dir_prefix);
   free(ftp_acc);
   free(ftp_pass);
   if (ftp_proxy)
      free(ftp_proxy);
   if (http_proxy)
      free(http_proxy);
   if (no_proxy)
      free(no_proxy);
}
