/* loop.c
   Module for handling iterating retrieve loops */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <errno.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif /* HAVE_STRING_H */
#include <ctype.h>
#ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
#else
#  include <time.h>
#endif
#endif
#include <assert.h>

#include "geturl.h"
#include "nmalloc.h"
#include "loop.h"
#include "recr.h"
#include "url.h"
#include "html.h"
#include "http.h"
#include "ftp.h"
#include "mtch.h"
#include "host.h"

/* global variables */
extern int verbose, quiet, kill_longer, kill_error, recursive;
extern int noclobber;
extern int always_rest, use_proxy;
extern int ftp_glob;
extern char *no_proxy;
extern int force_html;
extern char *fileprefix, *base_href, *ftp_acc, *ftp_pass;

extern FILE *lfile;

#ifdef DEBUG
extern int debug;
#endif

/* Show the dotted progress report of file loading. Called with length
   and a flag to tell it whether to reset or not. It keeps the offset
   information in static local variables. */
void
show_progress(long len, int res)
{
   static int offs, ndot, nrow;

   /* Length zero means initialization */
   if (len == 0)
   {
      offs = ndot = nrow = 0;
      fprintf(lfile, "\n%5dK ->",
	      nrow * DOTS_ON_LINE * BYTES_IN_DOT / 1024);
   }
   /* Offset gets incremented by current value. */
   offs += res;
   /* While offset is >= BYTES_IN_DOT, print dots, taking care that
      every 50th dot needs to be preceded by a status message. */
   for (; offs >= BYTES_IN_DOT; offs -= BYTES_IN_DOT)
   {
      if (!(ndot % DOT_SPACING))
	 fputc(' ', lfile);
      fputc('.', lfile);
      ++ndot;
      if (ndot == DOTS_ON_LINE)
	 ndot = 0;
      if (ndot == 0)
      {
	 ++nrow;
	 fprintf(lfile, "\n%5dK ->",
		 nrow * DOTS_ON_LINE * BYTES_IN_DOT / 1024);
      }
   }
}

/* The function returns pointer to a static char[] buffer in which
   zero-terminated string-representation of time (in form hh:mm:ss) is
   printed. */
char *
time_str(void)
{
   static char tms[15];
   time_t tim;
   struct tm *ptm;

   *tms = 0;
   tim = time(NULL);
   ptm = localtime(&tim);
   sprintf(tms, "%02d:%02d:%02d", ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
   return tms;
}

/* Retrieve the given URL. This should contain as few protocol -
   specific switches as possible, which should be chucked to
   protocol-specific functions (gethttp, getftp etc.). Right now it is
   quite far from that goal. Operations it performs are:

   1. Parse the URL
   2. Keep calling the appropriate function for the protocol, until
      the document is retrieved or the number of tries is exceeded.
   3. Print the error messages that lower-level functions won't
   handle. */
uerr_t
retrieve_url(char *url, char **file, int ntry, int *dt)
{
   FILE *fp;
   char *remote_dir, *remote_host, *newloc;
   int remote_port, count, res;
   long len, contlen, restval;
   uerr_t err, result, proto;
   static char http_error[256]; /* some servers print unbelievably
				   long messages! */
   char *tms;      /* Time string */
   struct stat st; /* Used by stat(2) */
   char *suf;      /* suffix, used for noclobber */
   char *proxy;
   char *acc, *passwd;
   int local_use_proxy;

   /* Optimize the URL! */
   url = opt_url(url, 0);
   /* Prepare memory for parseurl. */
   remote_dir = nmalloc(strlen(url) + 1);
   remote_host = nmalloc(strlen(url) + 1);
   /* Parse the URL. */
   proto = result = parseurl(url, remote_host, &remote_port, remote_dir);

   /* Decide whether a proxy is to be used. */
   local_use_proxy = (use_proxy && getproxy(proto)
		      && no_proxy_match(remote_host, no_proxy));
   
   /* If there are things to glob, call the specialized function to do
      it. */
   if (ftp_glob && urlproto(url) == URLFTP && has_wildcards(remote_dir)
       && !local_use_proxy)
      return ftp_glob_loop(url, ntry);
   
   switch (result)
   {
      case URLUNKNOWN:
	 if (!quiet)
	    fprintf(lfile, "Unknown/unsupported protocol.\n");
	 free(remote_dir);
	 free(remote_host);
	 free(url);
	 return URLUNKNOWN;
	 break;
      case URLBADPORT:
	 if (!quiet)
	    fprintf(lfile, "Invalid port specification.\n");
	 free(remote_dir);
	 free(remote_host);
	 free(url);
	 return URLBADPORT;
	 break;
      case URLBADHOST:
	 if (!quiet)
	    fprintf(lfile, "Invalid host name.\n");
	 free(remote_dir);
	 free(remote_host);
	 free(url);
	 return URLBADHOST;
	 break;
      case URLFTP:
	 err = parse_uname(url, &acc, &passwd);
	 if (!acc)
	 {
	    acc = nstrdup(ftp_acc);
	    passwd = nstrdup(ftp_pass);
	 }
	 else if (!passwd)
	    passwd = nstrdup("");
	 break;
   }
   if (local_use_proxy)
   {
      proxy = getproxy(proto);
      if (!proxy)
      {
	 if (!quiet)
	    fprintf(lfile, "Could not find proxy host.\n");
	 free(remote_dir);
	 free(remote_host);
	 free(url);
	 return PROXERR;
      }
      /* Proxy is actually HTTP. */
      proto = URLHTTP;
   }
   /* Reset the counter. */
   count = 0;

   /* Start the loop. */
   do
   {
      /* Increment the pass counter. */
      ++count;
      /* Suppose no restarting gets done. */
      restval = 0;
      /* Restarting will be employed only if three conditions are met:
	 1) Protocol is FTP
	 2) File exists
	 3) Pass count is greater than 1, all always_rest is set. */
      if ((count > 1 || always_rest) && proto == URLFTP && exists(*file))
	 if (stat(*file, &st) == 0)
	    restval = st.st_size;
      /* Get the current time string. */
      tms = time_str();
      /* Print fetch message, if verbose. */
      if (verbose)
      {
	 char *hurl = hide_url(url);
	 fprintf(lfile, "\n--%s--[%d]: %s\n                 \"%s\"\n",
		 tms, count, hurl, *file);
	 free(hurl);
      }
      /* If noclobber is turned on and file already exists, do not
         retrieve the file */
      if (noclobber && recursive && exists(*file) && !restval)
      {
	 if (verbose)
	    fprintf(lfile, "File already there, will not retrieve.\n");
	 /* If the file is there, we suppose it's retrieved OK. */
	 *dt |= RETROKF;
	 /* If its suffix is "html" or (yuck!) "htm", we suppose it's
            text/html, a harmless lie. */
	 if ((suf = suffix(*file))
	     && (!strcmp(suf, "html") || !strcmp(suf, "htm")))
	    *dt |= TEXTHTML;
	 free(suf);
	 free(remote_dir);
	 free(remote_host);
	 free(url);
	 if (proto == URLFTP)
	 {
	    free(acc);
	    free(passwd);
	 }
	 /* Another harmless lie: */
	 return RETROK;
      }
      /* Open the file to write to. */
      if (proto != URLFTP)
      {
	 fp = fopen(*file, "w");
	 if (!fp)
	 {
	    if (!quiet)
	       fprintf(lfile, "%s: %s\n", *file, mystrerror(errno));
	    free(remote_dir);
	    free(remote_host);
	    free(url);
	    return FOPENERR;
	 }
      }
      /* Default document type is empty. */
      *dt = 0;
      /* Call the appropriate routine, according to URL protocol and
	 local_use_proxy. */
      if (local_use_proxy)
	 err = proxy_retrieve(proxy, url, fp, &len, &contlen, &res,
			      &newloc, dt, http_error);
      else /* !local_use_proxy */
      {
	 if (proto == URLHTTP)
	    /* Retrieve the file through HTTP. */
	    err = gethttp(remote_host, remote_port, remote_dir,
			  fp, &len, &contlen, &res, &newloc, dt,
			  http_error);
	 else /* proto != URLHTTP */
	    err = getftp(remote_host, remote_port, remote_dir, *file,
			 acc, passwd, &len, restval, &res, NULL, dt);
      }
      /****
       * Retrieval is finished for now (whichever way), so we can
       * close the stream.
       ****/
      if (proto != URLFTP)
	 fclose(fp);
      if (kill_error && proto == URLHTTP)
	 if ((*dt & RETROKF) == 0)
	    unlink(*file);
      tms = time_str();
      if (proto == URLHTTP)  /* HTTP-specific stuff */
      {
	 switch (err)
	 {
	    case HERR: case CONTIMEDOUT: case CONSOCKERR:
	    case CONCLOSED: case READERR:
	       /* Non-fatal errors continue executing the loop, which
		  will bring them to "while" statement at the end, to
		  judge whether the number of tries was exceeded. */
	       continue;
	       break;
	    case HEOF: case HOSTERR: case CONREFUSED: case PROXERR:
	    case FOPENERR:
	       /* Fatal errors just return from the function. Note
		  that HEOF is one of the fatals, since it usually
		  means that request to the server was invalid for
		  some reason. If the connection was close while
		  reading headers, HERR would have been much more
		  probable. */
	       unlink(*file);
	       free(remote_host);
	       free(remote_dir);
	       free(url);
	       return err;
	       break;
	    case FWRITEERR:
	       /* Another fatal error. */
	       if (!quiet)
	       {
		  if (verbose)
		     fprintf(lfile, "\n");
		  fprintf(lfile, "Cannot write to \"%s\" (%s).\n",
			  *file, mystrerror(errno));
	       }
	       /* I will not unlink *file on purpose. */
	       free(remote_host);
	       free(remote_dir);
	       free(url);
	       return err;
	       break;
	    case NEWLOCATION:
	       /* A special kludge, if the location header was found,
		     to rename the file and restart retrieving. */
	       url = newloc;
	       free(remote_host);
	       free(remote_dir);
	       remote_dir = nmalloc(strlen(url));
	       remote_host = nmalloc(strlen(url));
	       result = parseurl(url, remote_host, &remote_port, remote_dir);
	       --count;
	       unlink(*file);
	       free(*file);
	       *file = url_filename(url, fileprefix);
	       continue;
	       break;
	    case RETRFINISHED:
	       /* Deal with you later. */
	       break;
	    default:
	       /* All possibilities should have been exhausted. */
	       assert(0);
	 }
      }
      else /* URLFTP */
      {
	 switch (err)
	 {
	    case CONSOCKERR: case CONTIMEDOUT: case CONERROR:
	    case FTPSRVERR: case FTPUNKNOWNTYPE:
	    case BINDERR: case CONPORTERR: case LISTENERR:
	    case FTPPORTERR: case ACCEPTERR:
	       /* Various non-fatal errors produce continuation. */
	       continue;
	       break;
	    case FTPLOGINC:
	       /* If login is incorrect, it can mean two things:
		  1. FTP server is too busy and does not allow logging
		  in, or
		  2. Login is incorrect.
		  If the number of tries is set to infinity, we must
		  take into account the possibility number two, and
		  set ntry to an arbitrary (presumably high, but not
		  infinite) value.  */
	       if (ntry == 0)
		  ntry = FTP_MAX_LINC;
	       continue;
	       break;
	    case HOSTERR: case CONREFUSED: case FTPNSFOD:
	    case FTPBADFILE:
	       /* Fatal errors -> return. */
	       unlink(*file);
	       free(remote_host);
	       free(remote_dir);
	       free(url);
	       free(acc);
	       free(passwd);
	       return err;
	       break;
	    case FWRITEERR:
	       /* Another fatal. */
	       if (!quiet)
		  fprintf(lfile, "Can't write to \"%s\" (%s).\n",
			  *file, mystrerror(errno));
	       /* I will not unlink *file on purpose. */
	       free(remote_host);
	       free(remote_dir);
	       free(url);
	       free(acc);
	       free(passwd);
	       return err;
	       break;
	    case FTPRETRINT: case RETRFINISHED: case FTPRERR:
	       /* Deal with you later. */
	       break;
	    default:
	       /* All possibilities should have been exhausted. */
	       assert(0);
	 }
      }
      if (proto == URLFTP) /* Again, FTP-specific stuff */
      {
	 if (err == RETRFINISHED)
	 {
	    if (verbose)
	       fprintf(lfile, "%s URL [%ld] -> \"%s\" [%d]\n",
		       tms, len, *file, count);
	    else if (!quiet)
	       fprintf(lfile, "%s URL: %s [%ld] -> \"%s\" [%d]\n",
		       tms, url, len, *file, count);
	    free(remote_host);
	    free(remote_dir);
	    free(url);
	    free(acc);
	    free(passwd);
	    return RETROK;
	 }
	 else /* It must be FTPRETRINT or FTPRERR, which boil down to
		 the same thing. */
	 {
	    if (verbose && len != 0)
	    {
	       fprintf(lfile, "\nData connection closed at byte %ld. ", len);
	       printwhat(count, ntry);
	    }
	    continue;
	 }
      }
      else /* proto == URLHTTP */
      {
	 /* If file got unlinked due to error, there is not much point
	    in telling the user it's there. */
	 if (kill_error)
	    if ((*dt & RETROKF) == 0)
	    {
	       if (!quiet)
		  fprintf(lfile, "%s %s, \"%s\" removed.\n", tms, http_error,
			  *file);
	    }
	 /* Checking for unlink only where the retrieval was
	       correct. */
	 if (len == contlen)
	 {
	    if (!kill_error || (*dt & RETROKF))
	    {
	       if (verbose)
		  fprintf(lfile, "%s URL [%ld/%ld] -> \"%s\" [%d]\n",
			  tms, len, contlen, *file, count);
	       else if (!quiet)
		  fprintf(lfile, "%s URL:%s [%ld/%ld] -> \"%s\" [%d]\n",
			  tms, url, len, contlen, *file, count);
	    }
	    free(remote_host);
	    free(remote_dir);
	    free(url);
	    return RETROK;
	 }
	 else if (res == 0) /* No read error */
	 {
	    if (contlen == -1) /* We don't know how much we
				  were supposed to get, so... */
	    {
	       if (!kill_error || (*dt & RETROKF))
	       {
		  if (verbose)
		     fprintf(lfile, "%s URL [%ld] -> \"%s\" [%d]\n",
			     tms, len, *file, count);
		  else if (!quiet)
		     fprintf(lfile, "%s URL:%s [%ld] -> \"%s\" [%d]\n",
			     tms, url, len, *file, count);
	       }
	       free(remote_host);
	       free(remote_dir);
	       free(url);
	       return RETROK;
	    }
	    else if (len < contlen) /* Now this means we lost the
				       connection too soon */
	    {
	       if (verbose)
	       {
		  fprintf(lfile, "Connection closed at byte %ld. ", len);
		  printwhat(count, ntry);
	       }
	       continue;
	    }
	    else if (!kill_longer) /* Meaning we got more than expected */
	    {
	       if (verbose)
		  fprintf(lfile, "%s URL [%ld/%ld] -> \"%s\" [%d]\n",
			  tms, len, contlen, *file, count);
	       else if (!quiet)
		  fprintf(lfile, "%s URL:%s [%ld/%ld] -> \"%s\" [%d]\n",
			  tms, url, len, contlen, *file, count);
	       free(remote_host);
	       free(remote_dir);
	       free(url);
	       return RETROK;
	    }
	    else /* The same, but not accepted */
	    {
	       if (verbose)
	       {
		  fprintf(lfile, "Connection closed at byte %ld/%ld. ",
			  len, contlen);
		  printwhat(count, ntry);
	       }
	       continue;
	    }
	 }
	 else /* Now res can only be -1 */
	 {
	    if (contlen == -1)
	    {
	       if (verbose)
	       {
		  fprintf(lfile, "Read error at byte %ld (%s).", len,
			  errno ? mystrerror(errno) : "timeout");
		  printwhat(count, ntry);
	       }
	       continue;
	    }
	    else /* res == -1 and contlen is given */
	    {
	       if (verbose)
	       {
		  fprintf(lfile, "Read error at byte %ld/%ld (%s). ",
			  len, contlen, errno ? mystrerror(errno) : "timeout");
		  printwhat(count, ntry);
	       }
	       continue;
	    }
	 }
	 /* not reached */
	 break;
      }
   } while (!ntry || (count < ntry));
   free(remote_host);
   free(remote_dir);
   free(url);
   if (proto == URLFTP)
   {
      free(acc);
      free(passwd);
   }
   return TRYLIMEXC;
}

/* Function retrieves the URL-s on the NULL-terminated list of char
   pointers, consecutively. Note that it calls recursive_retrieve, if
   necessary. */
int
retrieve_more_urls(char **url, int ntry)
{
   int allok;
   char *filename;
   int dt;

   /* Suppose all is OK. */
   allok = 1;
   filename = NULL;
   for (; *url; url++)
   {
      /* Construct the filename from URL name. */
      filename = url_filename(*url, fileprefix);
      /* Reset type. */
      dt = 0;
      /* Retrieve the URL. */
      if (retrieve_url(*url, &filename, ntry, &dt) != RETROK)
	 allok = 0;
      /* Call recursive retrieval if necessary. */
      if (recursive && ((dt & TEXTHTML) && (dt & RETROKF)))
	 recursive_retrieve(NULL, filename, ntry, 0, *url);
#ifdef DEBUG
      else if (recursive && debug)
	 fprintf(lfile, "%s is not text/html, so we don't chase\n", filename);
#endif
      if (filename)
	 free(filename);
   }
   return allok;
}

/* Retrieves URL-s in null-terminated list of pointers, then reads the
   file, and retrieves all the URL-s it can find there. Note that file
   need not be an HTML document, and will not be parsed as such (so,
   no tag searches etc). However, nothing prevents you to give it an
   HTML file -- all full URL-s (of the sort findurl from url.c can
   match) will be retrieved */
uerr_t
retrieve_urls_from_file(char **url, char *file, int ntry)
{
   long nread;
   FILE *fp;
   char *buf, *pbuf, *filename, *base, *current_url;
   int dt, size, dummy, no_proto;

   /* First retrieve the URL-s on the argument list. */
   if (url)
      retrieve_more_urls(url, ntry);
   if (strcmp(file, "-"))
   {
      fp = fopen(file, "r");
      if (!fp)
      {
	 if (!quiet)
	    fprintf(lfile, "%s: %s\n", file, mystrerror(errno));
	 return FOPENERR;
      }
   }
   else
      fp = stdin;
   /* Load the file. */
   load_file(fp, &buf, &nread);
   if (file)
      fclose(fp);
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "Loaded %s (size %ld).\n", file, nread);
#endif
   /* Find URL-s one by one, and retrieve them. Note that buffer is
      being changed in the process (URL-s get zero-terminated). */
   for (pbuf = buf;
	(pbuf = (force_html ? htmlfindurl(pbuf, nread - (pbuf - buf),
					  &size, file)
		 : findurl(pbuf, nread - (pbuf - buf), &size)));
	pbuf += size)
   {
      dt = 0;
      current_url = NULL;
      if (force_html)
      {
	 /* If a URL hasn't a protocol, findurl() won't be able to
	    find it, and will return another location or NULL. */
	 no_proto = (pbuf != findurl(pbuf, size, &dummy));
	 if (no_proto)
	 {
	    /* First, construct the base, which can be relative
	       itself. */
	    base = html_base(file);
	    if (!base)
	       base = base_href; /* This one can still be NULL... */
	    if (base)
	       current_url = construct(base, pbuf, size, no_proto);
	    else
	    {
	       if (!quiet)
	       {
		  char *temp = nmalloc(size + 1);
		  strncpy(temp, pbuf, size);
		  temp[size] = 0;
		  fprintf(lfile,
			  "Error: Relative link %s without a given base.\n",
			  temp);
		  free(temp);
	       }
	       continue;
	    }
	 }
      }
      if (!current_url)
      {
	 current_url = nmalloc(size + 1);
	 strncpy(current_url, pbuf, size);
	 current_url[size] = 0;
      }
      filename = url_filename(current_url, fileprefix);
      retrieve_url(current_url, &filename, ntry, &dt);
      free(current_url);
      free(filename);
   }
   /* Free the buffer. */
   free(buf);
   return RETROK;
}

/* Print 'giving up', or 'retrying', depending on the action to
   do. Numbers represent the attempt number and the attempt limit
   (please don't ask which one is which). */
void
printwhat(int n1, int n2)
{
   if (verbose)
   {
      if (n1 == n2)
	 fprintf(lfile, "Giving up.\n");
      else
	 fprintf(lfile, "Retrying.\n");
   }
}

	       
