/* recr.c
   Module for handling of recursive retrieving. */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <errno.h>
#include <assert.h>
#include <ctype.h>

#include "geturl.h"
#include "recr.h"
#include "nmalloc.h"
#include "loop.h"
#include "url.h"
#include "html.h"
#include "http.h"
#include "ftp.h"
#include "mtch.h"
#include "host.h"

/* global variables */
extern int verbose, quiet, recursive;
extern int maxreclevel, spanhost, relative_only;
extern int follow_ftp, simple_check, force_html, ntry;
extern char *no_proxy;
extern char *fileprefix, *base_href, *ftp_acc, *ftp_pass;
extern char *version_string;
extern int use_robots;

extern FILE *lfile;

#ifdef DEBUG
extern int debug;
#endif


/* The core of recursive retrieving.  Endless recursion is avoided by
   having all URL-s stored to a static hash table, so that none of
   them will be loaded twice. URL-s must often undergo changes (like
   stripping ../ and ./ in a reasonable way, rejecting hrefs to
   foreign hosts, or regulating suffix accepting/rejecting according
   to preset rules) for things to work properly.

   The function also supports maximum recursion level, and several
   more goodies. Coding is heavily recursive and possibly unstable. A
   good rewrite would be useful. */
uerr_t
recursive_retrieve(char **url, char *file, int ntry, int first_time,
		   char *this_url)
{
   /* The hash table of all undesirable-to-load
      URL-s: */
   static char **hash;
   static char **forbidden = NULL;
   static int level; /* Recursion level -- also static */
   
   long nread;
   FILE *fp;
   char *buf, *pbuf, *filename, *constr, *suf, *opt, *base;
   int dt, dummy, size, relative, no_proto, ftpurl;
   uerr_t err, proto;
   char *rurl, *rfile; /* For robots */

   /* Create the URL hash and retrieve URL-s from the argument line,
      if the first time. */
   if (first_time)
   {
      create_hash(&hash, url);
      retrieve_more_urls(url, ntry);
      level = 0;
      forbidden = NULL;
   }
   else
      ++level;

   /* Bail out if maxreclevel is exceeded. */
   if ((maxreclevel != 0) && (level > maxreclevel))
   {
#ifdef DEBUG
      if (debug)
	 fprintf(lfile, "Recursion level %d exceeded max. level %d.\n",
		 level, maxreclevel);
#endif
      --level;
      return RECLEVELEXC;
   }

   /* Get filename from which to read the URL-s (HTML).  If none is
      specified, take /dev/null, i.e. do nothing. */
   if (!file)
      file = "/dev/null";
   
   /* Load the file. */
   if (strcmp(file, "-"))
   {
      fp = fopen(file, "r");
      if (!fp)
      {
	 if (!quiet)
	    fprintf(lfile, "%s: %s\n", file, mystrerror(errno));
	 --level;
	 return FOPENERR;
      }
   }
   else
      fp = stdin;
   
   load_file(fp, &buf, &nread);
   
   if (file)
      fclose(fp);
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "Loaded %s (size %ld).\n", file, nread);
#endif

   /* The main loop: the idea is to call htmlfindurl and use constr()
      to construct the URL name. If the URL is of appropriate type,
      and it hasn't been loaded before, call the function recursively
      on that filename. */
   pbuf = buf;
   do
   {
      constr = NULL;
      /* If the function is called for the first time, use findurl()
	 instead of htmlfindurl(). */
      if (first_time && !force_html)
      {
	 pbuf = findurl(pbuf, nread - (pbuf - buf), &size);
	 /* If pbuf is NULL, exit the loop, else just copy the URL as
	    the construct to do with later. */
	 if (!pbuf)
	    break;
	 else
	 {
	    constr = nmalloc(size + 1);
	    strncpy(constr, pbuf, size);
	    constr[size] = 0;
	    no_proto = 0; /* As always with findurl. */
	    relative = 0;
	 }
      }
      else
      {
	 /* Call HTML parser to get the next appropriate URL. */
	 pbuf = htmlfindurl(pbuf, nread - (pbuf - buf), &size, file);
	 if (!pbuf)
	    break;
	 /* If a URL hasn't a protocol, findurl() won't be able to
	    find it, and will return another location or NULL. */
	 no_proto = (pbuf != findurl(pbuf, nread - (pbuf - buf), &dummy));
	 if (no_proto)
	 {
	    /* First, construct the base, which can be relative
	       itself.
	       
	       Criteria for creating the base are:
	       1) html_base created by <base href="...">
	       2) current URL
	       3) base provided from the command line */
	    base = NULL;
	    base = html_base(file);
	    if (!base)
	       base = this_url;
	    if (!base)
	       base = base_href;
	    if (!base)
	    {
	       if (!quiet)
	       {
		  char *temp = nmalloc(size + 1);
		  strncpy(temp, pbuf, size);
		  temp[size] = 0;
		  fprintf(lfile,
			  "Error: Relative link %s without a given base.\n",
			  temp);
		  free(temp);
	       }
	       continue;
	    }
	    if (this_url)
	       base = construct(this_url, base, strlen(base),
				base != findurl(base, strlen(base), &dummy));
	    else
	       base = nstrdup(base);
	    constr = construct(base, pbuf, size, no_proto);
	    free(base);
	 }
	 else /* has proto */
	 {
	    constr = nmalloc(size + 1);
	    strncpy(constr, pbuf, size);
	    constr[size] = 0;
	 }
	 /* A URL is relative if the host and protocol are not named,
	    and the name does not start with '/'. */
	 relative = no_proto && (*pbuf != '/');
#ifdef DEBUG
	 if (debug)
	    fprintf(lfile, "%s: this_url: %s\nname: %s\n", file, this_url, constr);
#endif
      }
      /* Optimize the URL without potential DNS lookup. */
      opt = opt_url(constr, 0);
      free(constr);
      constr = opt;
      /* Several checkings whether a file is acceptable to load:
	 1. check for protocol (either HTTP or FTP are acceptable)
	 2. check if URL is ftp, and we don't load it
	 3. check for relative links (if relative_only is set)
	 4. check for domain
	 5. check for suffix
	 6. check for same host (if spanhost is unset), with possible
	 gethostbyname baggage
	 7. check for robots.txt

	 Addendum: If the URL is FTP, and it is to be loaded, only the
	 domain and suffix settings are "stronger".
	 
	 Note that checkings are approximately in ascending order
	 of time consuming, and mutually excluding.
	 
	 Also note that .html and (yuck) .htm will get loaded
	 regardless of suffix rules (but that is remedied later with
	 unlink). */
      proto = urlproto(constr);
      ftpurl = (proto == URLFTP);
      /* If the proto is neither HTTP nor FTP, chuck it out. */
      if (!in_hash(hash, constr))
	 if (proto != URLHTTP && proto != URLFTP)
	    add_hash(&hash, constr);
      /* If it is FTP, and FTP is not followed, chuck it out. */
      if (!in_hash(hash, constr))
	 if (ftpurl && !follow_ftp)
	    add_hash(&hash, constr);
      /* If it is absolute link and they are not followed, chuck it
	 out. */
      if (!in_hash(hash, constr) && !ftpurl)
	 if (relative_only && !relative)
	    add_hash(&hash, constr);
      /* If its domain is not to be accepted/lookedup, chuck it
	 out. */
      if (!in_hash(hash, constr))
	 if (!accept_domain(constr))
	    add_hash(&hash, constr);
      /* If its suffix is not on the acceptance list, or is on the
	 rejection list, chuck it out. */
      if (!in_hash(hash, constr))
      {
	 if (!((suf = suffix(constr)) && (!strcmp(suf, "html")
					  || !strcmp(suf, "htm"))))
	    if (!acceptable(constr))
	       add_hash(&hash, constr);
	 if (suf)
	    free(suf);
      }
      /* Optimize the URL (which includes possible DNS lookup) only
	 after all other possibilities have been exhausted. */
      if (!in_hash(hash, constr))
      {
	 opt = opt_url(constr, !simple_check);
	 free(constr);
	 constr = opt;
	 if (!in_hash(hash, constr) && !ftpurl)
	    if (!spanhost && this_url && !same_host(this_url, constr))
	       add_hash(&hash, constr);
      }
      /* What about robots.txt? */
#ifdef USE_ROBOTS
      if (use_robots && this_url)
      {
	 if (!in_hash(hash, constr) && !ftpurl)
	 {
	    if (!tried_robots(this_url))
	    {
	       err = retrieve_robots(this_url, ROBOTS_FILENAME);
	       if (err == ROBOTSOK)
	       {
		  rurl = robots_url(this_url, ROBOTS_FILENAME);
		  rfile = url_filename(rurl, fileprefix);
		  forbidden = parse_robots(rfile);
		  free(rurl);
		  free(rfile);
	       }
	       else
		  forbidden = NULL;
	    }
   
	    if (!robots_match(constr, forbidden))
	    {
#ifdef DEBUG
	       if (debug)
		  fprintf(lfile, "Stuffing %s because %s forbids it.\n",
			  this_url, ROBOTS_FILENAME);
#endif
	       add_hash(&hash, constr);
	    }
	 }
      }
#endif /* USE_ROBOTS */

      filename = NULL;
      /* If it wasn't chucked out, do something with it. */
      if (!in_hash(hash, constr))
      {
	 /* Add it to the hash of already-loaded URL-s. */
	 add_hash(&hash, constr);
	 /* Reset its type. */
	 dt = 0;
	 /* Get the filename to save it to. */
	 filename = url_filename(constr, fileprefix);
	 /* Retrieve it. */
	 retrieve_url(constr, &filename, ntry, &dt);
	 /* If it is not en error, and it is of type text/html, parse
	    it recursively. */
	 if ((dt & TEXTHTML) && (dt & RETROKF))
	    recursive_retrieve(NULL, filename, ntry, 0, constr);
#ifdef DEBUG
	 else if (debug)
	    fprintf(lfile, "%s is not text/html so we don't chase.\n", filename);
#endif
	 /* If an suffix-rejected file was loaded only because it was HTML,
	    undo the error now */
	 if (!acceptable(constr))
	 {
	    if (verbose)
	       fprintf(lfile, "Killing %s since it should be rejected.\n",
		       filename);
	    if (unlink(filename) && !quiet)
	       fprintf(lfile, "unlink: %s\n", mystrerror(errno));
	 }
      }
#ifdef DEBUG
      else if (debug)
	 fprintf(lfile, "%s already in hash, so we don't load.\n", constr);
#endif
      /* Free filename and constr. */
      if (filename)
	 free(filename);
      free(constr);
      /* Increment the pbuf for the appropriate size. */
      pbuf += size;
   } while (pbuf);
   /* Free the buffer. */
   free(buf);
   /* Decrement the recursion level. */
   --level;
   return RETROK;
}


#ifdef USE_ROBOTS

/* Construct the robots URL. */
char *
robots_url(char *url, char *robots_filename)
{
   char *constr, *host, *dir;
   int port;
   int l = strlen(url);
   uerr_t err;

   host = nmalloc(l + 1);
   dir = nmalloc(l + 1);
   err = parseurl(url, host, &port, dir);
   assert(err == URLHTTP);
   constr = nmalloc(7 + strlen(host) + 1 + numdigit(port) + 1
		    + strlen(robots_filename) + 1);
   sprintf(constr, "http://%s:%d/%s", host, port, robots_filename);
   free(host);
   free(dir);
   return constr;
}

/* Retrieves the robots_filename from the root server directory, if
   possible. Returns ROBOTSOK if robots were retrieved OK, and
   NOROBOTS if robots could not be retrieved for any reason. */
uerr_t
retrieve_robots(char *url, char *robots_filename)
{
   int dt;
   char *constr, *file;
   uerr_t err;

   constr = robots_url(url, robots_filename);
   file = url_filename(constr, fileprefix);
   err = retrieve_url(constr, &file, ntry, &dt);
   free(constr);
   free(file);
   if (err == RETROK)
      return ROBOTSOK;
   else
      return NOROBOTS;
}

/* Parses the robots_filename returning the disallowed filenames in a
   malloc-ed vector of character pointers. */
char **
parse_robots(char *robots_filename)
{
   FILE *fp;
   char **entries;
   char *line, *cmd, *str;
   int c, len, num;

   entries = NULL;
   num = 0;
   fp = fopen(robots_filename, "r");
   if (!fp)
      return NULL;
   do
   {
      line = nmalloc(1);
      *line = 0;
      len = 0;
      /* Character loop to end of line: */
      do
      {
	 c = fgetc(fp);
	 /* Exit on EOF or newline. */
	 if (c == EOF || c == '\n')
	    break;
	 /* Increase the line size. */
	 line = nrealloc(line, len + 2);
	 line[len] = c;
	 line[len + 1] = 0;
	 ++len;
      } while (1);
      /* Destroy <CR> if there is one. */
      if (line[len] == '\r')
	 line[len] = 0;
      /* According to specifications, optional space may be at the
         end... */

#ifdef DEBUG
      if (debug)
	 fprintf(lfile, "Line: %s\n", line);
#endif
      /* Skip spaces. */
      for (cmd = line; *cmd && isspace(*cmd); cmd++);
      if (!*cmd)
      {
	 free(line);
	 continue;
      }
      /* Look for ':'. */
      for (str = cmd; *str && *str != ':'; str++);
      if (!*str)
      {
	 free(line);
	 continue;
      }
      /* Zero-terminate the command. */
      *str++ = 0;
      /* Look for the string beginning... */
      for (; *str && isspace(*str); str++);
      if (!*str)
      {
	 free(line);
	 continue;
      }
      if (!strcasecmp(cmd, "User-agent"))
      {
	 /* If Geturl is not matched, chuck the whole thing... */
	 if (fnmatch(str, version_string, 0) != 0)
	 {
	    int i;
	    free(line);
	    if (entries)
	    {
	       for (i = 0; entries[i]; i++)
		  free(entries[i]);
	       free(entries);
	    }
	    return NULL;
	 }
      }
      else if (!strcasecmp(cmd, "Disallow"))
      {
	 int i;
	 entries = (char **)nrealloc(entries, (num + 2)* sizeof(char *));
	 entries[num] = nstrdup(str);
	 entries[++num] = NULL;
	 /* Strip trailing spaces, according to specifications. */
	 for (i = strlen(str); i >= 0 && isspace(str[i]); i--)
	    if (isspace(str[i]))
	       str[i] = 0;
      }
      free(line);
   } while (c != EOF);
   return entries;
}

/* The functions keeps an internal database of which hosts have been
   quereied for robots.txt. */
int
tried_robots(char *url)
{
   static char **hosts = NULL;
   char **h, *host, *dir;
   int i, port, l = strlen(url);
   uerr_t err;

#ifdef DEBUG
   if (debug)
      fprintf(lfile, "Checking whether to load robots... ");
#endif
   host = nmalloc(l + 1);
   dir = nmalloc(l + 1);
   err = parseurl(url, host, &port, dir);
   assert(err == URLHTTP);
   free(dir);
   i = 0;
   if (hosts)
   {
      for (h = hosts; *h; h++, i++)
	 if (strcasecmp(host, *h) == 0)
	    break;
      if (*h)
      {
	 free(host);
#ifdef DEBUG
	 if (debug)
	    fprintf(lfile, "no\n");
#endif
	 return 1;
      }
   }
   hosts = nrealloc(hosts, (i + 2) * sizeof(char *));
   hosts[i] = nstrdup(host);
   hosts[i + 1] = NULL;
   free(host);
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "yes\n");
#endif
   return 0;
}

/* May the URL url be loaded according to disallowing rules stored in
   forbidden? */
int
robots_match(char *url, char **forbidden)
{
   char *host, *dir;
   int port, l;
   uerr_t err;

   if (!forbidden)
      return 1;
   l = strlen(url);
   host = nmalloc(l + 1);
   dir = nmalloc(l + 1);
   err = parseurl(url, host, &port, dir);
   if (err != URLHTTP)
   {
      free(host);
      free(dir);
      return 1; /* none of our business */
   }
   free(host);
   for (; *forbidden; forbidden++)
   {
#ifdef DEBUG
      if (debug)
	 fprintf(lfile, "Matching %s against %s... ", dir, *forbidden);
#endif
      l = strlen(*forbidden);
      /* If dir is forbidden, we may not load the file. */
      if (strncmp(dir, *forbidden, l) == 0)
      {
	 free(dir);
#ifdef DEBUG
	 if (debug)
	    fprintf(lfile, "matched.\n");
#endif
	 return 0; /* Matches, i.e. does not load... */
      }
   }
   free(dir);
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "not matched.\n");
#endif
   return 1;
}

#endif /* USE_ROBOTS */
