/* url.c
   Module for URL handling */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#include <assert.h>

#include "geturl.h"
#include "nmalloc.h"
#include "url.h"
#include "host.h"

extern int verbose, debug, prefix_files, simple_check;
extern int dirstruct, dirmode, add_hostdir, recursive, always_rest;
extern char *dir_prefix, *fileprefix;
extern FILE *lfile;

extern char **accepts, **rejects;
extern char *http_proxy, *ftp_proxy;

extern int errno;

/* NULL-terminated list of strings to be recognized as prototypes (URL
   schemes). Note that recognized doesn't mean supported -- only HTTP
   and FTP are supported for now.

   However, a string that does not match anything on the list will be
   considered a relative URL. */
char *protostrings[] = {
   "http://",
   "ftp://",
   "mailto:",
   "news:",
   "news://",
   "nntp://",
   "telnet://",
   "wais://",
   "file://",
   "gopher://",
   "prospero://",
   NULL
};

/* Similar to former, but for supported protocols: */
proto_t sup_protos[] = {
   { "http://", URLHTTP, DEFAULT_HTTP_PORT },
   { "ftp://", URLFTP, DEFAULT_FTP_PORT },
   { NULL, FTPOK, 0 }
};

/* Returns the number of characters to be skipped if the first thing
   in a URL is URL: (which is 0 or 4) */
int
skip_url(char *url)
{
   return strncasecmp(url, "URL:", 4) ? 0 : 4;
}

/* Returns 1 if the string contains unsafe characters, 0 otherwise. */
int
contains_unsafe(char *s)
{
   for (; *s; s++)
      if (strchr(URL_UNSAFE, *s))
	 return 1;
   return 0;
}

/* Decodes the forms %xx in a URL to the character the hexadecimal
   code of which is xx. xx are hexadecimal digits from
   [0123456789ABCDEF] or [0123456789abcdef]. */
void
decode_string(char *s)
{
   char *p = s;

   for (; *s; s++, p++)
   {
      if (*s != '%')
	 *p = *s;
      else
      {
	 if (!*(s + 1) || !*(s + 2))
	    break;
	 *p = (ASC2HEXD(*(s + 1)) << 4) + ASC2HEXD(*(s + 2));
	 s += 2;
      }
   }
   *p = 0;
}

/* Encodes the unsafe characters (listed in URL_UNSAFE) in a given
   string, returning a malloc-ed %XX encoded string. */
char *
encode_string(char *s)
{
   char *p, *res;
   int i;

   for (i = 0, p = s; *s; s++, i++)
      if (strchr(URL_UNSAFE, *s))
	 i += 2; /* Two more characters (hex digits) */
   res = nmalloc(i + 1);
   s = p;
   for (p = res; *s; s++)
      if (strchr(URL_UNSAFE, *s))
      {
	 *p++ = '%';
	 *p++ = HEXD2ASC(*s >> 4);
	 *p++ = HEXD2ASC(*s & 0xf);
      }
      else
	 *p++ = *s;
   *p = 0;
   return res;
}

/* Returns the proto-type if it is a supported protocol, or URLUNKNOWN
   if not. */
uerr_t
urlproto(char *url)
{
   int i;

   url += skip_url(url);
   for (i = 0; sup_protos[i].name; i++)
      if (!strncasecmp(url, sup_protos[i].name, strlen(sup_protos[i].name)))
	 return sup_protos[i].ind;
   for (i = 0; url[i] && url[i] != ':' && url[i] != '/'; i++);
   if (url[i] == ':')
   {
      for (++i; url[i] && url[i] != '/'; i++)
	 if (!isdigit(url[i]))
	    return URLBADPORT;
      if (url[i - 1] == ':')
	 return URLFTP;
      else
	 return URLHTTP;
   }
   else
      return URLHTTP;
}

/* Extract the given URL of the form (http:|ftp:)//hostname(:port)?/dirname
   1. hostname (terminated with '/' or ':')
   2. port number (terminated with '/'), or chosen for the protocol
   3. dirname (everything after hostname, '/' if hostname is unterminated
   Most errors are handled. No allocation is done, you must supply
   pointers to allocated memory.

   Later: Added recognizing of:
   hostname:/dir/file          for FTP and
   hostname(:portnum)/dir/file for HTTP.
   
   Note: hostname:dir/file, the form used by some clients to enable
   accessing the relative path to the server's, is not yet
   supported. */
uerr_t
parseurl(char *url, char *hostname, int *port, char *dirname)
{
   int i, l, recognizable, changed_port = 0;
   uerr_t type;

   url += skip_url(url);
   for (i = 0; protostrings[i]; i++)
      if (!strncasecmp(protostrings[i], url, strlen(protostrings[i])))
	 break;
   /* If proto is found, it is recognizable. */
   recognizable = protostrings[i] ? 1 : 0;
   for (i = 0; sup_protos[i].name; i++)
   {
      l = strlen(sup_protos[i].name);
      if (!strncasecmp(sup_protos[i].name, url, l))
         break;
   }
   /* If protocol is recognizable, but unsupported, bail out, else
      suppose unknown. */
   if (recognizable && !sup_protos[i].name)
      return URLUNKNOWN;
   else if (!sup_protos[i].name)
      type = URLUNKNOWN;
   else
      type = sup_protos[i].ind;
   
   if (type == URLUNKNOWN)
      l = 0;
   if (type != URLUNKNOWN)
   {
      /* Allow for a username and password to be added (i.e. just skip
         them). */
      char *p;
      
      for (p = url + l; *p && *p != '/'; p++)
	 if (*p == '@')
	    break;
      /* If a '@' was found before the first occurrance of '/', skip
         it. */
      if (*p == '@')
	 l += p - (url + l) + 1;
   }
   for (i = l; url[i] && url[i] != ':' && url[i] != '/'; i++)
      hostname[i - l] = url[i];
   hostname[i - l] = 0;
   if (!*hostname)
      return URLBADHOST;
   *port = 0;
   changed_port = 0;
   if (url[i] == ':')
   {
      changed_port = 1;
      for (++i; url[i] && url[i] != '/'; i++)
	 if (isdigit(url[i]))
	 {
	    *port *= 10;
	    *port += url[i] - '0';
	 }
	 else
	    return URLBADPORT;
      if (type == URLUNKNOWN)
      {
	 if (url[i - 1] == ':')
	 {
	    type = URLFTP;
	    changed_port = 0;
	 }
	 else
	    type = URLHTTP;
      }
      else if (url[i - 1] == ':')
	 return URLBADPORT;
   }
   else if (type == URLUNKNOWN)
      type = URLHTTP;
   if (!url[i])
      strcpy(dirname, "/");
   else
      strcpy(dirname, url + i);
   if (!*port)
   {
      if (changed_port)
	 return URLBADPORT;
      for (i = 0; sup_protos[i].name; i++)
	 if (sup_protos[i].ind == type)
	    break;
      if (!sup_protos[i].name)
	 return URLUNKNOWN;
      *port = sup_protos[i].port;
   }
   decode_string(hostname);
   decode_string(dirname);
   return type;
}

/* Find the optional username and password within the URL, as per
   RFC1738. */
uerr_t
parse_uname(char *url, char **user, char **passwd)
{
   int i;
   char *p, *col, **where;

   *user = NULL;
   *passwd = NULL;
   url += skip_url(url);
   /* Look for end of protocol string. */
   for (i = 0; protostrings[i]; i++)
      if (!strncasecmp(protostrings[i], url, strlen(protostrings[i])))
	 break;
   if (!protostrings[i])
      return URLUNKNOWN;
   /* Add protocol offset. */
   url += strlen(protostrings[i]);
   /* Is there an '@' sign? */
   for (p = url; *p && *p != '/'; p++)
      if (*p == '@')
	 break;
   /* If not, return. */
   if (*p != '@')
      return URLOK;
   /* Else find the username and password. */
   for (p = col = url; *p != '@'; p++)
   {
      if (*p == ':' && !*user)
      {
	 *user = nmalloc(p - url + 1);
	 memcpy(*user, url, p - url);
	 (*user)[p - url] = 0;
	 decode_string(*user);
	 col = p + 1;
      }
   }
   /* Decide whether you have only the username or both. */
   where = *user ? passwd : user;
   *where = nmalloc(p - col + 1);
   memcpy(*where, col, p - col);
   (*where)[p - col] = 0;
   decode_string(*where);
   return URLOK;
}

/* Hide the password field in a URL. */
char *
hide_url(char *url)
{
   char *res = nstrdup(url);
   char *p, *s;
   int i, pass;

   url += skip_url(url);
   /* Look for end of protocol string. */
   for (i = 0; protostrings[i]; i++)
      if (!strncasecmp(protostrings[i], url, strlen(protostrings[i])))
	 break;
   assert(protostrings[i]);
   /* Add protocol offset. */
   i = strlen(protostrings[i]);
   /* Search for '@' before the first '/'. */
   for (p = url + i; *p && *p != '/'; p++)
      if (*p == '@')
	 break;
   if (*p != '@')
      return res;
   pass = 0;
   for (p = url + i, s = res + i; *p ; p++, s++)
   {
      if (*p == '@' || *p == '/')
	 break;
      /* If at passwd, replace it with 'x'. */
      if (pass)
	 *s = 'x';
      else if (*p == ':')
	 pass = 1;
   }
   return res;
}

/* The function parses a modifiable buffer buf into directory part dir
   and filename part file. The dir and file are just pointers to parts
   of buf, and need not be freed */
void
parse_dir(char *buf, char **dir, char **file)
{
   int i;
   static char empty_dir[] = "/"; /* This is writable... */
   
   for (i = strlen(buf); i && buf[i] != '/'; i--);
   if (!i && buf[i + 1]) /* Just filename or /filename */
   {
      *dir = empty_dir;
      *file = buf + (*buf == '/' ? 1 : 0);
   }
   else if (!i) /* Just dirname */
   {
      *dir = buf;
      *file = "";
   }
   else /* Everything OK */
   {
      buf[i] = 0;
      *dir = buf;
      *file = buf + i + 1;
   }
}

/* Find URL of format
   scheme:hostname[:port]/dir -- no HTML
   parsing done, it can be any text/binary buffer */
char *
findurl(char *buf, int howmuch, int *count)
{
   char **prot, *s1, *s2;

   for (s1 = buf; howmuch; s1++, howmuch--)
      for (prot = protostrings; *prot; prot++)
	 if (howmuch <= strlen(*prot))
	    break;
	 else if (!strncasecmp(*prot, s1, strlen(*prot)))
	 {
	    for (s2 = s1, *count = 0;
		 *s2 && *s2 >= 32 && *s2 < 127 && !isspace(*s2) &&
		    !strchr(URL_SEPARATOR, *s2);
		 s2++, (*count)++);
	    return s1;
	 }
   return NULL;
}

/* 1. Parse the given URL, to extract hostname, dirname and filename
   2. Make the required hierarchy of directories, as with mkdir -p
   3. Return the pathname of the URL-equivalent filename */
char *
mkstruct(char *url)
{
   char *host, *nhost, *odir, *dir, *res;
   int port, pos, stat;
   uerr_t err;

   /* URL is like a box of chocolates -- you never know what you're
      gonna get. I always use "safe" malloc-ing. Maybe it would be
      better to make parseurl do the malloc, but it's complex enough
      as it is. :( */
   host = nmalloc(strlen(url) + 1);
   odir = nmalloc(strlen(url) + 1);
   err = parseurl(url, host, &port, odir);
   assert(err == URLHTTP || err == URLFTP);
   /* Add dir_prefix or hostname (whichever is required) to the
      beginning of dir */
   if (add_hostdir)
   {
      /* Checking for a host's true name (or at least a consistent
	 name for saving to directory) should be done here, reusing
	 the hhash, if possible. */
      if (!simple_check)
      {
	 nhost = realhost(host);
	 free(host);
	 host = nhost;
      }
      URL_CLEANSE(host);
      URL_CLEANSE(odir);
      dir = nmalloc(strlen(host) + strlen(odir) + 1);
      sprintf(dir, "%s%s", host, odir);
   }
   else
   {
      URL_CLEANSE(odir);
      dir = nmalloc(strlen(dir_prefix) + strlen(odir) + 1);
      sprintf(dir, "%s%s", dir_prefix, odir);
   }
   free(host);
   free(odir);
   for (pos = strlen(dir); pos && dir[pos] != '/'; pos--)
      if (dir[pos] == '?')
	 dir[pos] = 0; /* This takes care of CGI buggers */
   
   /* If dir is just a hostname, return index.html preceded with
      appropriate directory path. This is valid only if the URL type
      is URLHTTP. */
   if (!pos)
   {
      stat = mymkdir(dir);
      if (stat == -1)
      {
	 /* hmm */
      }
      res = nmalloc(strlen(dir) + 1 + 10 + 1);
      if (err == URLHTTP)
	 sprintf(res, "%s/index.html", dir);
      else
	 sprintf(res, "%s/", dir);
      return res;
   }
   else
   {
      dir[pos] = 0;
      stat = mymkdir(dir);
      if (stat == -1)
      {
	 /* hmm */
      }
      dir[pos] = '/';
      if (dir[pos + 1])
	 res = nstrdup(dir);
      else
      {
	 res = nmalloc(strlen(dir) + 10 + 1);
	 strcpy(res, dir);
	 if (err == URLHTTP)
	    strcat(res, "index.html");
      }
      free(dir);
      return res;
   }
}

/* The function that takes the dirname to be created, making sure that
   missing directories are made one by one. Its behaviour should be
   similar to mkdir -p. */
int
mymkdir(char *dir)
{
   int i, stat;

   /* Make a copy of dir, to be able to write to it. Otherwise, the
      function is unsafe if called with a const char *argument. */
   dir = nstrdup(dir);
   for (i = 0; 1; ++i)
   {
      for (; dir[i] && dir[i] != '/'; i++)
	 ;
      if (dir[i])
      {
	 dir[i] = 0;
	 stat = mkdir(dir, dirmode);
	 if (stat < 0 && errno != EEXIST)
	 {
	    free(dir);
	    return -1;
	 }
	 dir[i] = '/';
      }
      else
      {
	 stat = mkdir(dir, dirmode);
	 if (stat < 0 && errno != EEXIST)
	 {
	    free(dir);
	    return -1;
	 }
	 break;
      }
   }
   free(dir);
   return 0;
}

/* Create a unique filename, corresponding to a given URL.  If
   dirstruct is turned on, it will create all the necessary
   directories, and then return the full pathname of the file.  (it is
   important, since the function will not do what you want if you
   would like just to check for a filename) */
char *
url_filename(char *url, char *fileprefix)
{
   int count, i;
   char *name, *ptr, *murl, *pref, *dir, *file, *tmp;
   
   if (recursive && dirstruct)
   {
      char *url1;
      url1 = opt_url(url, 0);
      ptr = mkstruct(url1);
      free(url1);
      assert(ptr);
      tmp = nstrdup(ptr);
      parse_dir(tmp, &dir, &file);
      if (!*file)
      {
	 ptr = nrealloc(ptr, strlen(ptr) + strlen("index.html") + 1);
	 strcat(ptr, "index.html");
      }
      free(tmp);
      return ptr;
   }
   else if (prefix_files)
   {
      pref = nmalloc(strlen(dir_prefix) + 1 + strlen(fileprefix) + 1);
      sprintf(pref, "%s/%s", dir_prefix, fileprefix);
      for (count = 1; !(name = unique_name(pref, count)); count++)
	 ;
      free(pref);
      return name;
   }
   for (i = strlen(url); url[i] != '/' && i; i--)
      ;
   if (!i)
      return nstrdup(url);
   /* Checking for '~' is no longer necessary, since now it's
      done the "proper" way (through location hacks). However,
      I still provide checking for //hostname */
   if (url[i] == '/' && (!url[i + 1] || url[i - 1] == '/'))
      ptr = "index.html";
   else
      ptr = url + i + 1;
   murl = NULL;
   for (i = 0; ptr[i] && ptr[i] != '?'; i++)
      ;
   /* If there is a '?' in the string, destroy it and everything
      after it (takes care of CGI nonsense) */
   if (ptr[i] == '?')
   {
      murl = nstrdup(ptr);
      murl[i] = 0;
      ptr = murl;
   }
   pref = nmalloc(strlen(dir_prefix) + 1 + strlen(ptr) + 1);
   sprintf(pref, "%s/%s", dir_prefix, ptr);
   /* If we are dealing with FTP and always resting, no need for
      extensions. */
   if (urlproto(url) == URLFTP && always_rest)
   {
      if (murl)
	 free(murl);
      return pref;
   }
   for (count = 0; !(name = unique_name(pref, count)); count++)
      ;
   free(pref);
   if (murl)
      free(murl);
   return name;
}

/* How many digits in an integer? */
int
numdigit(int a)
{
   int res;

   for (res = 1; a /= 10; res++);
   return res;
}

/* Return a unique filename, given a prefix and count */
char *
unique_name(char *fileprefix, int count)
{
   char *filename;

   if (count)
   {
      filename = nmalloc(strlen(fileprefix) + numdigit(count) + 2);
      sprintf(filename, "%s.%d", fileprefix, count);
   }
   else
      filename = nstrdup(fileprefix);
   if (!exists(filename))
      return filename;
   else
   {
      free(filename);
      return NULL;
   }
}

/* Does a file exist? This is quite bad implementation, since it will
   return that a file does not exist if, e.g., the directory is
   unreadable. I don't mind it too much currently, though.  The proper
   way should, of course, be to check errno for ENOENT, but what about
   missing directories in the path? */
int
exists(char *filename)
{
   struct stat buf;
   return stat(filename, &buf) ? 0 : 1;
}

/* Construct an absolute URL, given a (possibly) relative one.  This
   is more tricky than it might seem, but it works. */
char *
construct(char *url, char *sub, int subsize, int no_proto)
{
   int i, fl;
   char *constr, *t;

   t = NULL;
   if (no_proto)
   {
      if (*sub != '/')
      {
	 for (i = strlen(url); i && url[i] != '/'; i--);
	 if (!i || (url[i] == url[i - 1]))
	 {
	    int l;
	    t = nmalloc((l = strlen(url)) + 2);
	    strcpy(t, url);
	    t[l] = '/';
	    t[l + 1] = 0;
	    url = t;
	    i = l;
	 }
	 constr = nmalloc(i + 1 + subsize + 1);
	 strncpy(constr, url, i + 1);
	 constr[i + 1] = 0;
	 strncat(constr, sub, subsize);
      }
      else
      {
	 i = 0;
	 do
	 {
	    for (; url[i] && url[i] != '/'; i++);
	    if (!url[i])
	       break;
	    if ((fl = (url[i] == url[i + 1] && url[i + 1] == '/')))
	       i += 2;
	 } while (fl);
	 if (!url[i])
	 {
	    int l;
	    t = nmalloc((l = strlen(url)) + 2);
	    strcpy(t, url);
	    t[l] = '/';
	    t[l + 1] = 0;
	    url = t;
	 }
	 constr = nmalloc(i + 1 + subsize + 1);
	 strncpy(constr, url, i);
	 constr[i] = 0;
	 strncat(constr + i, sub, subsize);
	 constr[i + subsize] = 0;
      }
   }
   else
   {
      constr = nmalloc(subsize + 1);
      strncpy(constr, sub, subsize);
      constr[subsize] = 0;
   }
   if (t)
      free(t);
   return constr;
}


/* Simplify absolute pathnames containing "." and ".." entries. Thus
   path_simplify("/a/b/c/./../d/..") returns "/a/b". The routine was
   taken from pdksh. */
void
path_simplify(char *path)
{
   char	*f;
   char	*cur = path;
   char	*t = path + 1;
   
   /* Only simplify absolute paths */
   if (*cur != '/')
      return;
   
   while (1)
   {
      /* treat multiple '/'s as one '/' */
      while (*t == '/')
	 t++;
      
      if (*t == 0)
      {
	 if (cur != path)
	    *cur = 0;
	 else
	    *(cur + 1) = 0;
	 break;
      }
      
      /* find/copy next component of pathname */
      *cur = '/';
      f = cur + 1;
      while (*t != 0 && *t != '/')
	 *f++ = *t++;
      
      /* check for a ". or ".." entry and simplify */
      if (*(cur + 1) == '.')
      {
	 if (*(cur + 2) == '.' && (cur + 3) == f)
	 {
	    if (cur != path)
	    {
	       cur -= 2;
	       while (*cur != '/')
		  cur--;
	    }
	 }
	 else if ((cur + 2) != f)
	    cur = f;
      }
      else
	 cur = f;
   }
}


/* URL optimization -- dot-cleans the URL, and, if flag is set, calls
   realname, and returns the malloc-ed URL with the new hostname. */
char *
opt_url(char *url, int fl)
{
   char *host, *dir, *tmp, *res, *dirpart, *filepart, *user, *passwd;
   int i, l, port;
   uerr_t err, proto;

   /* Parse the URL to see what we're dealing with. */
   host = nmalloc(strlen(url) + 1);
   dir = nmalloc(strlen(url) + 1);
   err = parseurl(url, host, &port, dir);
   parse_uname(url, &user, &passwd);
   if (err != URLHTTP && err != URLFTP)
   {
      free(host);
      free(dir);
      return nstrdup(url);
   }
   /* Parse the dir to see what's a directory and what's a file. */
   parse_dir(dir, &dirpart, &filepart);
   path_simplify(dirpart);
   if (fl)
   {
      tmp = realhost(host);
      free(host);
      host = tmp;
   }
   proto = urlproto(url);
   for (i = 0; sup_protos[i].name; i++)
      if (sup_protos[i].ind == proto)
	 break;
   URL_CLEANSE(host);
   if (user)
      URL_CLEANSE(user);
   if (passwd)
      URL_CLEANSE(passwd);
   dirpart = encode_string(dirpart);
   filepart = encode_string(filepart);
   l = strlen(sup_protos[i].name) + (user ? (strlen(user) + 1) : 0) +
      (passwd ? (strlen(passwd) + 1) : 0) + strlen(host) +
      strlen(dirpart) + strlen(filepart) + 15; /* safe sex */
   res = nmalloc(l);
   sprintf(res, "%s%s%s%s%s%s:%d%s%s%s", sup_protos[i].name,
	   (user ? user : ""), (passwd ? ":" : ""),
	   (passwd ? passwd : ""), (user ? "@" : ""),
	   host, port, dirpart,
	   (dirpart[strlen(dirpart) - 1] == '/' ? "" : "/"), filepart);
   free(host);
   free(dir);
   free(dirpart); /* These two because of CLEANSE-ing */
   free(filepart);
   return res;
}

/* Load file to memory, return the malloc-ed buffer, and the file size */
void
load_file(FILE *fp, char **buf, long *nread)
{
   *nread = 0;
   *buf = NULL;
   while (!feof(fp))
   {
      *buf = nrealloc(*buf, *nread + FILE_BUFFER_SIZE);
      *nread += fread(*buf + *nread, sizeof(char), FILE_BUFFER_SIZE, fp);
   }
}

/* Create hash table, generic */
void
create_hash(char ***hash, char **url)
{
   if (!*hash)
   {
      if (url)
      {
	 int i;
	 for (i = 0; url[i]; i++);
	 *hash = nmalloc((i + 1) * sizeof(char *));
	 for (i = 0; url[i]; i++)
	    (*hash)[i] = nstrdup(url[i]);
	 (*hash)[i] = NULL;
      }
      else
      {
	 *hash = nmalloc(sizeof(char *));
	 **hash = NULL;
      }
   }
}

/* Is there an entry in hash table (generic)? */
int
in_hash(char **hash, char *s)
{
   for (; *hash; hash++)
      if (!strcmp(s, *hash))
	 return 1;
   return 0;
}

/* Add an entry to the hash table (generic) */
void
add_hash(char ***hash, char *s)
{
   int i;
   
   for (i = 0; (*hash)[i]; i++);
   *hash = nrealloc(*hash, (i + 2) * sizeof(char *));
   (*hash)[i] = nstrdup(s);
   (*hash)[i + 1] = NULL;
#ifdef DEBUG
   if (debug)
      fprintf(lfile, "Stored %s as %s in hash table.\n", s, (*hash)[i]);
#endif
}

/* Determine whether a URL is acceptable to be followed,
   according to a list of suffixes to accept/reject */
int
acceptable(char *constr)
{
   if (accepts)
   {
      if (rejects)
	 return in_acclist(accepts, constr) && !in_acclist(rejects, constr);
      else
	 return in_acclist(accepts, constr);
   }
   else if (rejects)
      return !in_acclist(rejects, constr);
   return 1;
}

/* Is a suffix in acceptance list? */
int
in_acclist(char **accepts, char *constr)
{
   int res;
   char *suf;

   suf = suffix(constr);
   if (!suf)
      return 0;
   res = matches(accepts, suf);
   free(suf);
   return res;
}

/* Return the malloc-ed suffix of a filename */
char *
suffix(char *s)
{
   int i;
   
   for (i = strlen(s); i && s[i] != '/' && s[i] != '.'; i--);
   if (s[i++] == '.')
      return nstrdup(s + i);
   else
      return NULL;
}

/* Does a string match any element of the list? */
int
matches(char **list, char *what)
{
   int i;

   for (i = 0; list[i]; i++)
      if (!strcmp(list[i], what))
	 return 1;
   return 0;
}

/* Parse a string containing comma-separated elements, and return
   a vector of char pointers with the elements. The comma-separated
   list is space-insensitive. */
char **
sepstring(char *optstr)
{
   char *tok, **res, *tmp;
   int i, j, l;
   
   if (!optstr || !strlen(optstr))
      return NULL;
   /* Strip spaces from optstr. */
   l = strlen(optstr);
   tmp = nmalloc(l + 1);
   for (i = j = 0; i < l; i++)
      if (!isspace(optstr[i]))
	 tmp[j++] = optstr[i];
   tmp[j] = 0;
   optstr = tmp;
   if (!strchr(optstr, ','))
   {
      res = nmalloc(2 * sizeof(char *));
      res[0] = nstrdup(optstr);
      res[1] = NULL;
      free(optstr);
      return res;
   }
   tok = strtok(optstr, ",");
   res = nmalloc(2 * sizeof(char *));
   res[0] = nstrdup(tok);
   res[1] = NULL;
   for (i = 1; (tok = strtok(NULL, ",")); i++)
   {
      res = nrealloc(res, (i + 2) * sizeof(char *));
      res[i] = nstrdup(tok);
      res[i + 1] = NULL;
   }
   free(optstr);
   return res;
}

/* Free acceptance and rejectance lists */
void
freelists(void)
{
   int i;

   if (accepts)
   {
      for (i = 0; accepts[i]; i++)
	 free(accepts[i]);
      free(accepts);
   }
   if (rejects)
   {
      for (i = 0; rejects[i]; i++)
	 free(rejects[i]);
      free(rejects);
   }
}

/* Returns proxy host address, according to protocol. */
char *
getproxy(uerr_t proto)
{
   if (proto == URLHTTP)
      return http_proxy ? http_proxy : getenv("http_proxy");
   else if (proto == URLFTP)
      return ftp_proxy ? ftp_proxy : getenv("ftp_proxy");
   else
      return NULL;
}

/* Should a host be accessed through proxy, concerning no_proxy? */
int
no_proxy_match(char *host, char *no_proxy)
{
   char **dom;
   int i, res;

   if (!no_proxy)
      return 1;
   if (!(dom = sepstring(no_proxy)))
      return 1;
   res = !sufmatch(dom, host);
   for (i = 0; dom[i]; i++)
      free(dom[i]);
   free(dom);
   return res;
}
